************************************************************************
!
! User element for vehicle bridge interaction basing on Yang's formula
! Yang Y.B., Wu Y.S. A versatile element for analyzing vehicle bridge
! interaction response. Engineering Structures, 2001, 23: 452-469.
! 
! Hossein Azimi, Khaled Galal, Oscar Pekau. A modified numerical VBI
! element for vehicles with constant velocity including road irregularities.
! Engineering Geology, 2011. This paper is refered for including the road
! irregularties into the VBI formulation. May 27, 2019 by SL @ ZJUT

! Both 2D and 3D 2-node E-B beam element are formulated here.
! The Hermite interpolation function is adopted.

! Nodal variables are the displacements and the rotations.
!
! This subroutine is for the following element types
!  > two-dimensional 2 node E-B beam elements with axial effects.
!  > three-dimensional 2 node E-B beam elements with axial effects.
!
! The stiffness, mass and damping matrix of the E-B element can be
! given explicitly. As a result, no numerical integration will be needed.
! However, the transformation should be performed to transform the above
! matrix that are formed in local coordinated to global coordinates.
!
! No mechanical or traction- and pressure-type boundary conditions
! are going to be applied to this VBI element, since they can be 
! given through the dummy elements inherent to ABAQUS. The element contains a 
! suspended vehicle carriage (rigid mass supported by combination of 
! stiffness and damper) and a rigid wheel mass. 
! Moreover, the irregularities between the wheel and the road/rail is considered. 
! However, the contact stiffness is not included, i.e. the wheel is in direct contact
! with the road. In other words, the contact stiffness is infinite, since in the 
! deformation compatibility equation there is no term associating with spring compression,
! i.e. (P(t)-W)/kH, P(t) is dynamic wheel/road force, kH is stiffness of contact spring.
! 

! Li Shi @ ZJUT, April 18, 2019 

***********************************************************************
!
! User element statement in the input file (set ? values as needed):
!
!  2D elements,finished on April 23, 2019, (Unsymmetric solver should be used)
!  *User Element,Nodes=2,Type=U2,Iproperties=1,Properties=47+40,Coordinates=2,Variables=44
!  1,2,6,USYM
!
!  3D elements,finished on June 23, 2019, (Unsymmetric solver should be used)
!  *User Element,Nodes=2,Type=U3,Iproperties=1,Properties=49+40,Coordinates=3,Variables=80
!  1,2,3,4,5,6,USYM
!
!
!     State Variables
!     --------------------------------------------------------------
!     G|t and G|t-deltat should be stored as internal variables, as they
!     contain contributions from the beam, wheel and vehicle. 
!     In the input file 'Variables=44' for 2D VBI element
!     In the input file 'Variables=80' for 3D VBI element

!     Material Properties Vector for 2D VBI element, max 8 wheels can be considered
!     modified on Nov 11, 2019
!     --------------------------------------------------------------
!     rhob   = props(1) ! mass density of beam
!     Areab  = props(2) ! section area of beam
!     EIb    = props(3) ! bending rigidity of beam
!     EAb    = Props(4) ! Axial tension stiffness of beam
!     vmass(16)  = props(5~20) ! mass of vehicle carriage (modeled as rigid mass of one dof)
!     wmass(16)  = props(21~36) ! mass of vehicle wheel (modeled as rigid mass of one dof)
!     susk(16)   = props(37~52) ! stiffness of vehicle suspension
!     susc(16)   = props(53~68) ! damping of vehicle suspension
!     x0(16)     = props(69~84) ! Initial position (t=0) for the frontmost wheel mass
!     velo   = props(85) ! Speed of vehicle
!     raydamp_a = props(86) ! Rayleigh damping factor of beam, proportional to mass
!     raydamp_b  = props(87) ! Rayleigh damping factor of beam, proportional to stiffness
!     nc     = jprops(1) ! Number of wheels, nc<=4
!***********************************************************************
!     Material Properties Vector for 3D VBI element
!     --------------------------------------------------------------
!     Local coordinate system of the E-B beam element is defined as
!     This definition is taken from <Programming the Finite Element Method, 4th Edition>
!     by I.M. Smith and D.V. Griffiths, and this coordinate system does not obey the right-rand rule.
!                   ^y'
!                   |    
!                   |        
!                   |
!                   |  
!                   |
!                   ---------------->x'
!                    \
!                     \
!                      \
!                       \
!                        \ z'
!
!     x', y' and z' are local coordinates associating with this E-B element, where x' denotes the long
!     axis of the element, y' and z' are coordinates defined on cross section of the beam. For beam of 
!     rectangular section, the y'-z' coordinates are defined as follows, i.e. y' and z' are along the
!     height and width of the rectangular cross section.
!                   ^y'
!                   |        
!                   |
!                   |  
!                   |   
!           -----------------
!           |               |
! z'<-------|               |
!           |               |
!           -----------------
!
!
!
!     Suppose 1. the wheel is moving along positive x' axle of the E-B beam element
!             2. the wheel and the suspended vehicle mass is acted along the negative y' axle
!     --------------------------------------------------------------
!     rhob   = props(1) ! mass density of beam
!     Areab  = props(2) ! section area of beam
!     EIby   = props(3) ! bending rigidity of beam about y axle, i.e. bending in x-z plane
!     EIbz   = props(4) ! bending rigidity of beam about z axle, i.e. bending in x-y plane
!     EAbx   = Props(5) ! Axial tension stiffness of beam along x axle
!     GJbx   = Props(6) ! Torsional stiffness of beam about x axle
!     vmass(16)  = props(7~22) ! mass of vehicle carriage (modeled as rigid mass of one dof)
!     wmass(16)  = props(23~38) ! mass of vehicle wheel (modeled as rigid mass of one dof)
!     susk(16)   = props(39~54) ! stiffness of vehicle suspension
!     susc(16)   = props(55~70) ! damping of vehicle suspension
!     x0(16)     = props(71~86) ! Initial position (t=0) for the frontmost wheel mass
!     velo   = props(87) ! Speed of vehicle, a negative value should be assigned if the direction 
!                          of moving is not along the positive local x axle.
!     raydamp_a = props(88) ! Rayleigh damping factor of beam, proportional to mass
!     raydamp_b  = props(89) ! Rayleigh damping factor of beam, proportional to stiffness
!     nc     = jprops(1) ! Number of wheels, nc<=4, maximum wheel number is 4
!***********************************************************************


****************************************************************************

      SUBROUTINE UEL(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +     PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +     TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +     PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +     NJPROP,PERIOD)

*
      IMPLICIT NONE
*
*     VARIABLES DEFINED IN UEL, PASSED BACK TO ABAQUS
*
      REAL(8) :: RHS,AMATRX,SVARS,ENERGY
*
*     VARIABLES PASSED INTO UEL 
*
      REAL(8) :: PROPS,coords,Uall,DUall,Vel,Accn,TIME,
     1  DTIME,PARAMS,ADLMAG,PREDEF,DDLMAG,PNEWDT,PERIOD
      INTEGER :: NDOFEL,NRHS,NSVARS,NPROPS,MCRD,NNODE,JTYPE,KSTEP,KINC,
     1  JELEM,NDLOAD,JDLTYP,NPREDF,LFLAGS,MLVARX,MDLOAD,JPROPS,NJPROP
*
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),coords(MCRD,NNODE),Uall(NDOFEL),
     2 DUall(MLVARX,*),Vel(NDOFEL),Accn(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)

      integer lenJobName,lenOutDir,nDim
      character*256 jobName,outDir,fileName





      !----------------------------------------------------------------
      ! 
      ! Perform initial checks
      !
      !
      ! Open the debug/error message file
      !
      call getJobName(jobName,lenJobName)
      call getOutDir(outDir,lenOutDir)
      fileName = outDir(1:lenOutDir)//'\aaMSGS_'//
     +     jobName(1:lenJobName)//'.dat'
      open(unit=80,file=fileName,status='unknown')


      ! Check the procedure type, this should be an implicit direct
      ! integration procedure,
      !  which are any of the following (11,12)
      !
      if((lflags(1).eq.11).or.(lflags(1).eq.12)) then
         !
         ! all is good
         !
      else
         write(*,*) 'Abaqus does not have the right procedure'
         write(*,*) 'go back and chekc the procedure type'
         write(*,*) 'lflags(1)=',lflags(1)
         write(80,*) 'Abaqus does not have the right procedure'
         write(80,*) 'go back and chekc the procedure type'
         write(80,*) 'lflags(1)=',lflags(1)
         call xit
      endif


      ! Check to see if you are doing a general
      !  step or a linear purturbation step
      !
      if(lflags(4).eq.1) then
         !
         ! lflags(4)=0 -> general step
         ! lflags(4)=1 -> linear purturbation step
         !
         write(*,*) 'Abaqus thinks you are doing'
         write(*,*) 'a linear purturbation step'
         write(80,*) 'Abaqus thinks you are doing'
         write(80,*) 'a linear purturbation step'
         call xit         
      endif


      ! Do nothing if a ``dummy'' step
      !
      if(dtime.eq.0.0) return
      !
      ! Done with initial checks
      !
      !----------------------------------------------------------------



      !----------------------------------------------------------------
      ! 
      ! Call the paricular element to perform the analysis
      !
      if(jtype.eq.2) then
         !
         ! This is a 2-D plane strain analysis, two nodes E-B beam element
         !
         nDim = 2
         call UPE2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +        PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +        TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +        PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +        NJPROP,PERIOD,nDim)
         
      elseif(jtype.eq.3) then
         !
         ! This is a 3D analysis, two nodes E-B beam element
         !
         nDim = 3
         call U3D2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +        PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +        TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +        PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +        NJPROP,PERIOD,nDim)
         !
         !
      else
         !
         ! We have a problem...
         !
         write(*,*) 'Element type not supported, jtype=',jtype
         write(80,*) 'Element type not supported, jtype=',jtype
         call xit
         !
      endif
      !
      ! Done with this element, RHS and AMATRX already returned
      !  as output from the specific element routine called
      !
      !----------------------------------------------------------------


      return
      end subroutine uel

************************************************************************

      subroutine UPE2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +     PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +     TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +     PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +     NJPROP,PERIOD,nDim)

*
      IMPLICIT NONE
*
*     VARIABLES DEFINED IN UEL, PASSED BACK TO ABAQUS
*
      REAL(8) :: RHS,AMATRX,SVARS,ENERGY
*
*     VARIABLES PASSED INTO UEL 
*
      REAL(8) :: PROPS,coords,Uall,DUall,Vel,Accn,TIME,
     1  DTIME,PARAMS,ADLMAG,PREDEF,DDLMAG,PNEWDT,PERIOD
      INTEGER :: NDOFEL,NRHS,NSVARS,NPROPS,MCRD,NNODE,JTYPE,KSTEP,KINC,
     1  JELEM,NDLOAD,JDLTYP,NPREDF,LFLAGS,MLVARX,MDLOAD,JPROPS,NJPROP,
     1  ndim
*
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),coords(MCRD,NNODE),Uall(NDOFEL),
     2 DUall(MLVARX,*),Vel(NDOFEL),Accn(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)

      real*8 vmass(16),wmass(16),susk(16),susc(16),rhob,areab,Eib,raydamp_a,
     +       raydamp_b,x0(16),velo,Eab,eln,x1x,x2x,wgh,
     +       xn,xc,xmax,xmin,stf(ndofel,ndofel),xksi,shph(ndofel,1),
     +       phuu,phwu,mslp(ndofel,ndofel),a0,a1,a2,a3,a4,a5,a6,a7,
     +       alpha,beta,gama,rdamp(ndofel,ndofel),ydx(ndofel),
     +       gcrt(ndofel),glst(ndofel),fxt(ndofel),cntf,wdisp,wvel,wacc,
     +       vinc,vfc,velt,qucrt,qwcrt,qulst,qwlst,qcrt(ndofel),
     +       qlst(ndofel)
      real*8 shcn(1,ndofel), shcn1d(1,ndofel), 
     *       shcn2d(1,ndofel), bps, bps1d, bps2d

      integer*4 nc,i,iflag,j,iter
      
      real*8 zero,one,two,half,Pi,three,third,gacc
      parameter(zero=0.d0,one=1.d0,two=2.d0,half=0.5d0,Pi=3.141592653d0,
     +     three=3.d0,third=1.d0/3.d0,gacc=-9.81d0)
      
      ! Obtain material properties
      rhob=props(1)
      areab=props(2)
      Eib=props(3)
      Eab=props(4)
      
      vmass(1)=props(5)
      vmass(2)=props(6)
      vmass(3)=props(7)
      vmass(4)=props(8)
      vmass(5)=props(9)
      vmass(6)=props(10)
      vmass(7)=props(11)
      vmass(8)=props(12)
      vmass(9)=props(13)
      vmass(10)=props(14)
      vmass(11)=props(15)
      vmass(12)=props(16)
      vmass(13)=props(17)
      vmass(14)=props(18)
      vmass(15)=props(19)
      vmass(16)=props(20)
      
      wmass(1)=props(21)
      wmass(2)=props(22)
      wmass(3)=props(23)
      wmass(4)=props(24)
      wmass(5)=props(25)
      wmass(6)=props(26)
      wmass(7)=props(27)
      wmass(8)=props(28)
      wmass(9)=props(29)
      wmass(10)=props(30)
      wmass(11)=props(31)
      wmass(12)=props(32)
      wmass(13)=props(33)
      wmass(14)=props(34)
      wmass(15)=props(35)
      wmass(16)=props(36)
      
      susk(1)=props(37)
      susk(2)=props(38)
      susk(3)=props(39)
      susk(4)=props(40)
      susk(5)=props(41)
      susk(6)=props(42)
      susk(7)=props(43)
      susk(8)=props(44)
      susk(9)=props(45)
      susk(10)=props(46)
      susk(11)=props(47)
      susk(12)=props(48)
      susk(13)=props(49)
      susk(14)=props(50)
      susk(15)=props(51)
      susk(16)=props(52)
      
      susc(1)=props(53)
      susc(2)=props(54)
      susc(3)=props(55)
      susc(4)=props(56)
      susc(5)=props(57)
      susc(6)=props(58)
      susc(7)=props(59)
      susc(8)=props(60)
      susc(9)=props(61)
      susc(10)=props(62)
      susc(11)=props(63)
      susc(12)=props(64)
      susc(13)=props(65)
      susc(14)=props(66)
      susc(15)=props(67)
      susc(16)=props(68)
      
      x0(1)=props(69)
      x0(2)=props(70)
      x0(3)=props(71)
      x0(4)=props(72)
      x0(5)=props(73)
      x0(6)=props(74)
      x0(7)=props(75)
      x0(8)=props(76)
      x0(9)=props(77)
      x0(10)=props(78)
      x0(11)=props(79)
      x0(12)=props(80)
      x0(13)=props(81)
      x0(14)=props(82)
      x0(15)=props(83)
      x0(16)=props(84)
      
      velo=props(85)
      raydamp_a=props(86)
      raydamp_b=props(87)
      
      nc=jprops(1)
      if (nc .gt. 16) then
          write(*,*) 'The number of wheels should <=16, nc=',nc
          write(80,*) 'The number of wheels should <=16, nc=',nc
          call xit
      endif
      
      ! Obtain time integration parameters, used in condensing the vehicle
      ! dofs into those of the wheel basing on the Newmark scheme
      alpha=params(1)
      beta=params(2)
      gama=params(3)
      a0=one/beta/dtime/dtime
      a1=one/beta/dtime
      a2=one/two/beta-one
      a3=dtime*(one-gama)
      a4=gama*dtime
      a5=a4*a0
      a6=a4*a1-one
      a7=a4*a2-a3

      ! Suppose the wheel is moving along x axle
      ! suppose spacing between neighboring wheels are larger than
      ! the length of beam element, so that a given beam element will
      ! be loaded by ONLY one wheel.
      x1x=dabs(coords(1,1))
      x2x=dabs(coords(1,2))
      xmax=max(x1x,x2x)
      xmin=min(x1x,x2x)
      eln=xmax-xmin
      if (eln .le. 0.0d0) then
          write(*,*) 'Zero element length, pls check ...'
          write(80,*) 'Zero element length, pls check ...'
          call xit
      endif
      iflag=0
      iter=0
      do i=1,nc,1
          xn=x0(i)
          xc=xn+velo*time(2)
          if (abs(xc).le.xmax .and. abs(xc).gt.xmin) then
              ! This conventional beam element will be transformed
              ! into VBI element
              iter=i
              iflag=1
              xksi=(abs(xc)-xmin)/(xmax-xmin)
              !write(80,*) xc,xmin,xksi
              exit
          endif
      enddo
      
      
              
      ! Form stiffness matrix, upper half
      stf=zero
      stf(1,1)=Eab/eln
      stf(1,4)=-Eab/eln
      stf(2,2)=12.0d0*Eib/(eln**3)
      stf(2,3)=6.0d0*Eib/(eln**2)
      stf(2,5)=-12.0d0*Eib/(eln**3)
      stf(2,6)=6.0d0*Eib/(eln**2)
      stf(3,3)=4.0d0*Eib/eln
      stf(3,5)=-6.0d0*Eib/(eln**2)
      stf(3,6)=2.0d0*Eib/eln
      stf(4,4)=Eab/eln
      stf(5,5)=12.0d0*Eib/(eln**3)
      stf(5,6)=-6.0d0*Eib/(eln**2)
      stf(6,6)=4.0d0*Eib/eln
      
      ! Form mass matrix, lumped mass is used, upper half
      mslp=zero
      mslp(1,1)=rhob*areab*eln/two
      mslp(2,2)=rhob*areab*eln/two
      mslp(4,4)=rhob*areab*eln/two
      mslp(5,5)=rhob*areab*eln/two
      
      ! Form mass matrix, consist mass is used, upper half
      !wgh=rhob*areab*eln
      !mslp=zero
      !mslp(1,1)=third*wgh
      !mslp(1,4)=third*half*wgh
      !mslp(2,2)=156.0d0*wgh/420.0d0
      !mslp(2,3)=-22.0d0*eln*wgh/420.0d0
      !mslp(2,5)=54.0d0*wgh/420.0d0
      !mslp(2,6)=13.0d0*eln*wgh/420.0d0
      !mslp(3,3)=4.0d0*(eln**2)*wgh/420.0d0
      !mslp(3,5)=-13.0d0*eln*wgh/420.0d0
      !mslp(3,6)=-three*(eln**2)*wgh/420.0d0
      !mslp(4,4)=third*wgh
      !mslp(5,5)=156.0d0*wgh/420.0d0
      !mslp(5,6)=22.0d0*eln*wgh/420.0d0
      !mslp(6,6)=4.0d0*(eln**2)*wgh/420.0d0
      
      ! Form stiffness and mass matrix, lower half, use symmetry
      do i=1,6,1
          do j=i+1,6,1
              stf(j,i)=stf(i,j)
              mslp(j,i)=mslp(i,j)
          enddo
      enddo
              
      
      ! Form damp matrix, Rayleigh damping is assumed.
      rdamp=raydamp_a*mslp+raydamp_b*stf
      
      ! Hermite interpolation for determing contribution
      ! of moving vehicle on beam stiffness matrix
      if (iflag .eq. 1) then
          
          ! Hermite interpolation func at current local coord xksi
          shph=zero
          shph(2,1)=one-three*xksi**2+two*xksi**3
          shph(5,1)=three*xksi**2-two*xksi**3
          
          ! Consistent interpolation
          shcn=zero
          shcn(1,2)=shph(2,1)
          shcn(1,5)=shph(5,1)
          shcn(1,3)=xksi*(one-two*xksi+xksi**2)*eln
          shcn(1,6)=xksi*(xksi**2-xksi)*eln
          
          ! First order derivative of shcn with respect to xksi
          shcn1d=zero
          shcn1d(1,2)=two*three*xksi*(xksi-one)
          shcn1d(1,5)=two*three*xksi*(one-xksi)
          shcn1d(1,3)=(one-two*two*xksi+three*xksi**2)*eln
          shcn1d(1,6)=(three*xksi**2-two*xksi)*eln
          
          ! Second order derivative of shcn with respect to xksi
          shcn2d=zero
          shcn2d(1,2)=two*three*(two*xksi-one)
          shcn2d(1,5)=two*three*(one-two*xksi)
          shcn2d(1,3)=(three*xksi-two)*two*eln
          shcn2d(1,6)=(three*xksi-one)*two*eln
          
          
          ! Update stiffness and damping matrix for VBI element
          phuu=a0*vmass(iter)+a5*susc(iter)+susk(iter)
          phwu=-a5*susc(iter)-susk(iter)
          
          stf=stf+(one+phwu/phuu)*susk(iter)*matmul(shph,shcn)
          stf=stf+(wmass(iter)*(velo/eln)**2)*matmul(shph,shcn2d)
          stf=stf+(one+phwu/phuu)*susc(iter)*(velo/eln)*
     +            matmul(shph,shcn1d)

          rdamp=rdamp+(one+phwu/phuu)*susc(iter)*matmul(shph,shcn)
          rdamp=rdamp+(two*wmass(iter)*velo/eln)*matmul(shph,shcn1d)
          
          ! Update mass matrix for VBI element
          mslp=mslp+wmass(iter)*matmul(shph,shcn)

      endif
      ! Forming AMATRX
      AMATRX=a0*mslp+(one+alpha)*a5*rdamp+(one+alpha)*stf
          
      ! SVARS(NSVARS) contains contribution of beam disp, beam velocity
      ! external force on beam, and contribution of moving vehicle only if
      ! the current beamt element is a VBI element, at time t+deltat and t
      ! for internal variables at time t
      ! i.e. SVARS(1:6)=YDX|t=-C*Vel-K*Uall,svars(7:12)=FXT|t;svars(13:18)=q|t;
      !        svars(19:21)={du,duvel,duacc}|t; svars(22)=flag|t;
      ! for internal variables at time t-deltat
      ! i.e. SVARS(22+1:22+6)=YDX|t-deltat,svars(22+7:22+12)=FXT|t-deltat; &
      !      svars(22+13:22+18)=q|t-deltat,svars(22+19:22+21)={du,duvel,duacc}|t-deltat; &
      !      svars(22+22)=flag|t-deltat
      ! i.e. SVARS(NSVARS=44)
      ! G|t+deltat=-C*Vel|t+deltat-K*Uall|t+deltta+Fxt|t+deltat-svars(13:18)
      ! G|t=svars(1:6)+svars(7:12)-svars(35:40)
      ! RHS|t+deltat=-M*Acc|t+deltat+(1+alpha)*G|t+deltat-alpha*G|t
      
      gcrt=zero
      glst=zero
      fxt=zero
      ydx=zero
      
      ydx=-matmul(rdamp,Vel)-matmul(stf,Uall)
      glst=svars(1:6)
      gcrt=ydx
      if (iflag .eq. 1) then
          
          ! xc - coordinate of the wheel at global coord sys
          ! bps: amplitude of road roughness at xc
          ! bps1d: the first derivative of road roughness at xc
          ! bps2d: the second derivative of road roughness at xc
          xc=x0(iter)+velo*time(2)
          call roadrough (xc,bps,bps1d,bps2d)
          
          do i=1,ndofel,1
              fxt(i)=shph(i,1)*(vmass(iter)+wmass(iter))*gacc
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*susk(iter)*bps
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*
     +                      susc(iter)*bps1d*velo
              fxt(i)=fxt(i)-shph(i,1)*wmass(iter)*bps2d*velo*velo
          enddo
          gcrt=gcrt+fxt

          if (int(svars(22)) .eq. 1) then
              gcrt=gcrt-svars(13:18)
              glst=glst+svars(7:12)
          endif
          
          if (int(svars(44)) .eq. 1) then
              glst=glst-svars(35:40)
          endif
          
          svars(29:44)=svars(7:22)
          svars(7:12)=fxt
          
          ! Obtain disp, vel and acc at contact point for t+deltat
          wdisp=zero
          wvel=zero
          wacc=zero
          do i=1,ndofel,1
              wdisp=wdisp+shcn(1,i)*Uall(i)
              wvel=wvel+shcn(1,i)*Vel(i)+velo*shcn1d(1,i)*Uall(i)/eln
              wacc=wacc+shcn(1,i)*Accn(i)+
     *                  two*velo*shcn1d(1,i)*Vel(i)/eln+
     *                  shcn2d(1,i)*Uall(i)*(velo/eln)**2
          enddo
          ! Include contribution of road roughness into wheel response
          wdisp=wdisp+bps
          wvel=wvel+velo*bps1d
          wacc=wacc+velo*velo*bps2d
          
          cntf=-susc(iter)*wvel-susk(iter)*wdisp
          vfc=vmass(iter)*(a1*svars(20)+a2*svars(21))+ 
     +        susc(iter)*(a6*svars(20)+a7*svars(21))-
     +        susk(iter)*svars(19)
          vinc=(vfc-cntf)/phuu
          svars(19)=svars(19)+vinc
          velt=svars(20)
          svars(20)=a5*vinc-a6*velt-a7*svars(21)
          svars(21)=a0*vinc-a1*velt-a2*svars(21)
          svars(22)=real(iflag)
          
          qucrt=vmass(iter)*(a1*svars(20)+a2*svars(21))+
     +          susc(iter)*(a6*svars(20)+a7*svars(21))-
     +          susk(iter)*svars(19)
          qwcrt=-susc(iter)*(a6*svars(20)+a7*svars(21))+
     +          susk(iter)*svars(19)
          do i=1,ndofel,1
              qcrt(i)=shph(i,1)*(phwu/phuu*qucrt-qwcrt)
          enddo
          svars(13:18)=qcrt
          
      endif
      rhs(1:ndofel,1)=-matmul(mslp,Accn)+(one+alpha)*gcrt-alpha*glst
      
      ! Update internal variable at the end of current increment
      svars(23:28)=svars(1:6)
      svars(1:6)=ydx
      
      ! Update energy, energy(1): kinematic energy, 
      ! energy(2): elastic strain energy
      energy(1)=zero
      do i=1,2,1
          energy(1)=energy(1)+half*Vel(i)*(half*rhob*areab*eln)*Vel(i)
      enddo
      do i=4,5,1
          energy(1)=energy(1)+half*Vel(i)*(half*rhob*areab*eln)*Vel(i)
      enddo
      energy(2)=zero
      energy(2)=energy(2)+half*Eab*(Uall(1)-Uall(4))**2/eln
      energy(2)=energy(2)+half*Eib*12.0d0*(Uall(2)-Uall(5))**2/eln**3

      return
      end subroutine UPE2

************************************************************************
      subroutine U3D2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +        PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +        TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +        PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +        NJPROP,PERIOD,nDim)
*
      IMPLICIT NONE
*
*     VARIABLES DEFINED IN UEL, PASSED BACK TO ABAQUS
*
      REAL(8) :: RHS,AMATRX,SVARS,ENERGY
*
*     VARIABLES PASSED INTO UEL 
*
      REAL(8) :: PROPS,coords,Uall,DUall,Vel,Accn,TIME,
     1  DTIME,PARAMS,ADLMAG,PREDEF,DDLMAG,PNEWDT,PERIOD
      INTEGER :: NDOFEL,NRHS,NSVARS,NPROPS,MCRD,NNODE,JTYPE,KSTEP,KINC,
     1  JELEM,NDLOAD,JDLTYP,NPREDF,LFLAGS,MLVARX,MDLOAD,JPROPS,NJPROP,
     1  ndim
*
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),coords(MCRD,NNODE),Uall(NDOFEL),
     2 DUall(MLVARX,*),Vel(NDOFEL),Accn(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)
      
      real*8 vmass(2),wmass(2),susk(2),susc(2),rhob,areab,Eiby,Eibz,
     +       raydamp_a,raydamp_b,x0(2),velo,Eabx,GJbx,eln,x1x,x2x,
     +       wgh,xn,xc,xmax,xmin,stf(ndofel,ndofel),xksi,shph(ndofel,1),
     +       phuu,phwu,mslp(ndofel,ndofel),a0,a1,a2,a3,a4,a5,a6,a7,
     +       alpha,beta,gama,rdamp(ndofel,ndofel),ydx(ndofel),
     +       gcrt(ndofel),glst(ndofel),fxt(ndofel),cntf,wdisp,wvel,wacc,
     +       vinc,vfc,velt,qucrt,qwcrt,qulst,qwlst,qcrt(ndofel),
     +       qlst(ndofel),coordjb(ndim,nnode),ujb(ndofel),vjb(ndofel),
     +       ajb(ndofel)
      real*8 shcn(1,ndofel), shcn1d(1,ndofel), 
     *       shcn2d(1,ndofel), bps, bps1d, bps2d
      real*8 b1,b2,b3,b4,b5,b6,b7,b8

      integer*4 nc,i,iflag,j,iter
      
      real*8 zero,one,two,half,Pi,three,third,gacc
      parameter(zero=0.d0,one=1.d0,two=2.d0,half=0.5d0,Pi=3.141592653d0,
     +     three=3.d0,third=1.d0/3.d0,gacc=-9.81d0)
      
      ! Obtain material properties
      rhob=props(1)
      areab=props(2)
      Eiby=props(3)
      Eibz=props(4)
      Eabx=props(5)
      GJbx=props(6)
      
      vmass(1)=props(7)
      vmass(2)=props(8)
      wmass(1)=props(15)
      wmass(2)=props(16)
      susk(1)=props(23)
      susk(2)=props(24)
      susc(1)=props(31)
      susc(2)=props(32)
      x0(1)=props(39)
      x0(2)=props(40)
      velo=props(47)
      raydamp_a=props(48)
      raydamp_b=props(49)
      
      nc=jprops(1)
      if (nc .gt. 16) then
          write(*,*) 'The number of wheels should <=16, nc=',nc
          write(80,*) 'The number of wheels should <=16, nc=',nc
          call xit
      endif

      ! Obtain time integration parameters, used in condensing the vehicle
      ! dofs into those of the wheel basing on the Newmark scheme
      alpha=params(1)
      beta=params(2)
      gama=params(3)
      a0=one/beta/dtime/dtime
      a1=one/beta/dtime
      a2=one/two/beta-one
      a3=dtime*(one-gama)
      a4=gama*dtime
      a5=a4*a0
      a6=a4*a1-one
      a7=a4*a2-a3
      
      ! suppose spacing between neighboring vehicles are larger than
      ! the length of beam element, so that a given beam element will
      ! be loaded by ONLY one vehicle.
      call cordtran(mcrd,ndim,nnode,coords,coordjb)
      x1x=dabs(coordjb(1,1))
      x2x=dabs(coordjb(1,2))
      xmax=max(x1x,x2x)
      xmin=min(x1x,x2x)
      eln=xmax-xmin
      
      if (eln .le. 0.0d0 .or. x2x .lt. x1x) then
          write(80,*) 'Zero element length, pls check ...'
          call xit
      endif
      
      iflag=0
      iter=0
      do i=1,nc,1
          xn=x0(i)
          xc=xn+velo*time(2)
          if (xc.le.xmax .and. xc.gt.xmin) then
              ! This conventional beam element will be transformed
              ! into VBI element
              iter=i
              iflag=1
              xksi=(xc-xmin)/eln
              exit
          endif
      enddo
      
      ! Form stiffness matrix in elemental local coordinates
      ! taken from the book <Programming the Finite Element Method, 4th Edition>
      ! by I.M. Smith and D.V. Griffiths.
      b1=Eabx/eln
      b2=12.0d0*Eibz/(eln**3)
      b3=12.0d0*Eiby/(eln**3)
      b4=6.0d0*Eibz/(eln**2)
      b5=6.0d0*Eiby/(eln**2)
      b6=4.0d0*Eibz/eln
      b7=4.0d0*Eiby/eln
      b8=GJbx/eln
      
      stf=zero
      stf(1,1)=b1
      stf(7,7)=b1
      stf(1,7)=-b1
      stf(7,1)=-b1
      stf(2,2)=b2
      stf(8,8)=b2
      stf(2,8)=-b2
      stf(8,2)=-b2
      stf(3,3)=b3
      stf(9,9)=b3
      stf(3,9)=-b3
      stf(9,3)=-b3
      stf(4,4)=b8
      stf(10,10)=b8
      stf(4,10)=-b8
      stf(10,4)=-b8
      stf(5,5)=b7
      stf(11,11)=b7
      stf(5,11)=half*b7
      stf(11,5)=half*b7
      stf(6,6)=b6
      stf(12,12)=b6
      stf(6,12)=half*b6
      stf(12,6)=half*b6
      stf(2,6)=b4
      stf(6,2)=b4
      stf(2,12)=b4
      stf(12,2)=b4
      stf(6,8)=-b4
      stf(8,6)=-b4
      stf(8,12)=-b4
      stf(12,8)=-b4
      stf(5,9)=b5
      stf(9,5)=b5
      stf(9,11)=b5
      stf(11,9)=b5
      stf(3,5)=-b5
      stf(5,3)=-b5
      stf(3,11)=-b5
      stf(11,3)=-b5
      
      ! Form mass matrix in elemental local coordinates
      ! lumped mass matrix, zero mass warning has been issued by ABAQUS
      wgh=rhob*areab*eln
      mslp=zero
      mslp(1,1)=wgh/two
      mslp(2,2)=wgh/two
      mslp(3,3)=wgh/two
      mslp(4,4)=wgh/two
      mslp(7,7)=wgh/two
      mslp(8,8)=wgh/two
      mslp(9,9)=wgh/two
      mslp(10,10)=wgh/two
      
      ! Form mass matrix in elemental local coordinates
      ! consistent mass matrix
      !wgh=rhob*areab*eln
      !mslp=zero
      !mslp(1,1)=wgh/3.0d0
      !mslp(7,7)=wgh/3.0d0
      !mslp(1,7)=wgh/6.0d0
      !mslp(7,1)=wgh/6.0d0
      !mslp(2,2)=156.0d0*wgh/420.0d0
      !mslp(8,8)=156.0d0*wgh/420.0d0
      !mslp(2,8)=54.0d0*wgh/420.0d0
      !mslp(8,2)=54.0d0*wgh/420.0d0
      !mslp(3,3)=156.0d0*wgh/420.0d0
      !mslp(9,9)=156.0d0*wgh/420.0d0
      !mslp(3,9)=54.0d0*wgh/420.0d0
      !mslp(9,3)=54.0d0*wgh/420.0d0
      !mslp(4,4)=wgh/3.0d0
      !mslp(10,10)=wgh/3.0d0
      !mslp(4,10)=wgh/6.0d0
      !mslp(10,4)=wgh/6.0d0
      !mslp(5,5)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(11,11)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(5,11)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(11,5)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(6,6)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(12,12)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(6,12)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(12,6)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(2,6)=22.0d0*eln*wgh/420.0d0
      !mslp(6,2)=22.0d0*eln*wgh/420.0d0
      !mslp(2,12)=-13.0d0*eln*wgh/420.0d0
      !mslp(12,2)=-13.0d0*eln*wgh/420.0d0
      !mslp(6,8)=13.0d0*eln*wgh/420.0d0
      !mslp(8,6)=13.0d0*eln*wgh/420.0d0
      !mslp(8,12)=-22.0d0*eln*wgh/420.0d0
      !mslp(12,8)=-22.0d0*eln*wgh/420.0d0
      !mslp(5,9)=13.0d0*eln*wgh/420.0d0
      !mslp(9,5)=13.0d0*eln*wgh/420.0d0
      !mslp(9,11)=-22.0d0*eln*wgh/420.0d0
      !mslp(11,9)=-22.0d0*eln*wgh/420.0d0
      !mslp(3,5)=22.0d0*eln*wgh/420.0d0
      !mslp(5,3)=22.0d0*eln*wgh/420.0d0
      !mslp(3,11)=-13.0d0*eln*wgh/420.0d0
      !mslp(11,3)=-13.0d0*eln*wgh/420.0d0
      
      ! Form damp matrix, Rayleigh damping is assumed.
      rdamp=raydamp_a*mslp+raydamp_b*stf        
              
      ! Hermite interpolation for determing contribution
      ! of moving vehicle on beam stiffness matrix
      if (iflag .eq. 1) then
          
          ! Hermite interpolation func at current local coord xksi
          shph=zero
          shph(2,1)=one-three*xksi**2+two*xksi**3
          shph(8,1)=three*xksi**2-two*xksi**3
          
          ! Consistent interpolation
          shcn=zero
          shcn(1,2)=shph(2,1)
          shcn(1,8)=shph(8,1)
          shcn(1,6)=xksi*(one-two*xksi+xksi**2)*eln
          shcn(1,12)=xksi*(xksi**2-xksi)*eln
          
          ! First order derivative of shcn with respect to xksi
          shcn1d=zero
          shcn1d(1,2)=two*three*xksi*(xksi-one)
          shcn1d(1,8)=two*three*xksi*(one-xksi)
          shcn1d(1,6)=(one-two*two*xksi+three*xksi**2)*eln
          shcn1d(1,12)=(three*xksi**2-two*xksi)*eln
          
          ! Second order derivative of shcn with respect to xksi
          shcn2d=zero
          shcn2d(1,2)=two*three*(two*xksi-one)
          shcn2d(1,8)=two*three*(one-two*xksi)
          shcn2d(1,6)=(three*xksi-two)*two*eln
          shcn2d(1,12)=(three*xksi-one)*two*eln
          
          
          ! Update stiffness and damping matrix for VBI element
          phuu=a0*vmass(iter)+a5*susc(iter)+susk(iter)
          phwu=-a5*susc(iter)-susk(iter)
          
          stf=stf+(one+phwu/phuu)*susk(iter)*matmul(shph,shcn)
          stf=stf+(wmass(iter)*(velo/eln)**2)*matmul(shph,shcn2d)
          stf=stf+(one+phwu/phuu)*susc(iter)*(velo/eln)*
     *            matmul(shph,shcn1d)

          rdamp=rdamp+(one+phwu/phuu)*susc(iter)*matmul(shph,shcn)
          rdamp=rdamp+(two*wmass(iter)*velo/eln)*matmul(shph,shcn1d)
          
          ! Update mass matrix for VBI element
          mslp=mslp+wmass(iter)*matmul(shph,shcn)

      endif
      ! Forming AMATRX
      AMATRX=a0*mslp+(one+alpha)*a5*rdamp+(one+alpha)*stf
      
      
      
      ! SVARS(NSVARS) contains contribution of beam disp, beam velocity
      ! external force on beam, and contribution of moving vehicle only if
      ! the current beamt element is a VBI element, at time t+deltat and t
      ! for internal variables at time t
      ! i.e. SVARS(1:12)=YDX|t=-C*Vel-K*Uall,svars(13:24)=FXT|t;svars(25:36)=q|t;
      !        svars(37:39)={du,duvel,duacc}|t; svars(40)=flag|t;
      ! for internal variables at time t-deltat
      ! i.e. SVARS(40+1:40+12)=YDX|t-deltat,svars(40+13:40+24)=FXT|t-deltat; &
      !      svars(40+25:40+36)=q|t-deltat,svars(40+37:40+39)={du,duvel,duacc}|t-deltat; &
      !      svars(40+40)=flag|t-deltat
      ! i.e. SVARS(NSVARS=80)
      ! G|t+deltat=-C*Vel|t+deltat-K*Uall|t+deltta+Fxt|t+deltat-svars(25:36)
      ! G|t=svars(1:12)+svars(13:24)-svars(65:76)
      ! RHS|t+deltat=-M*Acc|t+deltat+(1+alpha)*G|t+deltat-alpha*G|t
      
      gcrt=zero
      glst=zero
      fxt=zero
      ydx=zero
      
      call eledoftran (mcrd,nnode,coords,NDOFEL,Uall,Vel,
     *                     Accn,ujb,vjb,ajb)
      ydx=-matmul(rdamp,vjb)-matmul(stf,ujb)
      glst=svars(1:12)
      gcrt=ydx
      if (iflag .eq. 1) then
          
          ! xc - coordinate of the wheel at local coord sys
          ! bps: amplitude of road roughness at xc
          ! bps1d: the first derivative of road roughness at xc
          ! bps2d: the second derivative of road roughness at xc
          xc=x0(iter)+velo*time(2)
          call roadrough (xc,bps,bps1d,bps2d)
          
          do i=1,ndofel,1
              fxt(i)=shph(i,1)*(vmass(iter)+wmass(iter))*gacc
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*susk(iter)*bps
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*
     *                      susc(iter)*bps1d*velo
              fxt(i)=fxt(i)-shph(i,1)*wmass(iter)*bps2d*velo*velo
          enddo
          gcrt=gcrt+fxt

          if (int(svars(40)) .eq. 1) then
              gcrt=gcrt-svars(25:36)
              glst=glst+svars(13:24)
          endif
          
          if (int(svars(80)) .eq. 1) then
              glst=glst-svars(65:76)
          endif
          
          svars(53:80)=svars(13:40)
          svars(13:24)=fxt
          
          ! Obtain disp, vel and acc at contact point for t+deltat
          wdisp=zero
          wvel=zero
          wacc=zero
          call eledoftran (mcrd,nnode,coords,NDOFEL,Uall,Vel,
     *                     Accn,ujb,vjb,ajb)
          do i=1,ndofel,1
              wdisp=wdisp+shcn(1,i)*ujb(i)
              wvel=wvel+shcn(1,i)*vjb(i)+velo*shcn1d(1,i)*ujb(i)/eln
              wacc=wacc+shcn(1,i)*ajb(i)+
     *                  two*velo*shcn1d(1,i)*vjb(i)/eln+
     *                  shcn2d(1,i)*ujb(i)*(velo/eln)**2
          enddo
          ! Include contribution of road roughness into wheel response
          wdisp=wdisp+bps
          wvel=wvel+velo*bps1d
          wacc=wacc+velo*velo*bps2d
          
          cntf=-susc(iter)*wvel-susk(iter)*wdisp
          vfc=vmass(iter)*(a1*svars(38)+a2*svars(39))+ 
     +        susc(iter)*(a6*svars(38)+a7*svars(39))-
     +        susk(iter)*svars(37)
          vinc=(vfc-cntf)/phuu
          svars(37)=svars(37)+vinc
          velt=svars(38)
          svars(38)=a5*vinc-a6*velt-a7*svars(39)
          svars(39)=a0*vinc-a1*velt-a2*svars(39)
          svars(40)=real(iflag)
          
          qucrt=vmass(iter)*(a1*svars(38)+a2*svars(39))+
     +          susc(iter)*(a6*svars(38)+a7*svars(39))-
     +          susk(iter)*svars(37)
          qwcrt=-susc(iter)*(a6*svars(38)+a7*svars(39))+
     +          susk(iter)*svars(37)
          do i=1,ndofel,1
              qcrt(i)=shph(i,1)*(phwu/phuu*qucrt-qwcrt)
          enddo
          svars(25:36)=qcrt
          
      endif
      rhs(1:ndofel,1)=-matmul(mslp,ajb)+(one+alpha)*gcrt-alpha*glst
      
      ! Transfer amatrx and rhs that are formed in local coordinates into global coordinates
      call amarhstran (mcrd,nnode,coords,MLVARX,NDOFEL,AMATRX,rhs)
      
      ! Update internal variable at the end of current increment
      svars(41:52)=svars(1:12)
      svars(1:12)=ydx
      
      ! Update energy, energy(1): kinematic energy, 
      ! energy(2): elastic strain energy
      energy(1)=zero
      do i=1,4,1
          energy(1)=energy(1)+half*vjb(i)*(wgh/two)*vjb(i)
      enddo
      do i=7,10,1
          energy(1)=energy(1)+half*vjb(i)*(wgh/two)*vjb(i)
      enddo
      energy(2)=zero
      ! for strain energy contributed by axle deformation along x axle
      energy(2)=energy(2)+half*Eabx*(ujb(1)-ujb(7))**2/eln
      ! for strain energy contributed by torsional deformation about x axle
      energy(2)=energy(2)+half*GJbx*(ujb(4)-ujb(10))**2/eln
      ! for strain energy contributed by bending in x-z plane
      energy(2)=energy(2)+half*Eiby*12.0d0*(ujb(3)-ujb(9))**2/eln**3
      energy(2)=energy(2)+half*Eiby*4.0d0*((eln*ujb(5))**2+
     +          (eln*ujb(11))**2+eln*eln*ujb(5)*ujb(11))/eln**3
      energy(2)=energy(2)+half*Eiby*12.0d0*(eln*ujb(5)+
     +          eln*ujb(11))*(ujb(3)-ujb(9))/eln**3
      ! for strain energy contributed by bending in y-x plane
      energy(2)=energy(2)+half*Eiby*12.0d0*(ujb(2)-ujb(8))**2/eln**3
      energy(2)=energy(2)+half*Eiby*4.0d0*((eln*ujb(6))**2+
     +          (eln*ujb(12))**2+eln*eln*ujb(6)*ujb(12))/eln**3
      energy(2)=energy(2)+half*Eiby*12.0d0*(eln*ujb(6)+
     +          eln*ujb(12))*(ujb(2)-ujb(8))/eln**3
  
      return
      
      end subroutine U3D2
************************************************************************
      subroutine roadrough (xc, bps, bps1d, bps2d)
      
!     ******************************************** 
!     road roughness of sine wave r(x)=rough*Sin(2*Pi*x/wvln+phag) is introduced, May 27, 2019
!      rough: amplitude of road roughness, modeled as a sine wave, m
!      wvln: props(16) ! wavelength of road roughness, modeled as a sine wave, m
!      phag: props(17) ! phase angle of road roughness, modeled as a sine wave, rad
!     ********************************************
!     random roughness is introduced, June 12, 2019, using methods given by Au et al
!     Au et al. Impact study of cable-stayed railway bridges with random rail irregularities,
!               Engineering Structures, 2002, 24: 529-541.
!       Parameters involved in this method are:
!       amp: Parameter of the psd function, unit: m^3
!       wn1: Parameter of the psd function, unit: 1/m
!       wn2: Parameter of the psd function, unit: 1/m
!       wnu: upper limit of the wavenumber, unit: 1/m
!       wnl: lower limit of the wavenumber, unit: 1/m
!       nsp: Number of sampling points, sufficiently large integer 
      IMPLICIT NONE
      real*8 xc, bps, bps1d, bps2d
      real*8 rough, phag, wvln, two, pi, af, lamdaf, lf, a0f, anf, bnf
      real*8 amp, wn1, wn2, wnu, wnl, wnk, wnstep, psd, ak
      integer nsp, ii, nf
      real*8 fq(500)
      
      ! bumper used in field test in Yanhe Bridge, Lianshui, Jiangsu
      ! af: height of the bumper, m
      ! lamdaf: width of the bumper, m
      ! lf: spacing of the bumper, m
      ! nf: number of Fourier terms approximating the bumper
      !pi=3.141592654d0
      !af=0.03d0
      !lamdaf=0.15d0
      !lf=1.39d0
      !nf=1000
    
      !a0f=af*lamdaf/pi/lf  
      !bps=a0f
      !bps1d=0.0d0
      !bps2d=0.0d0
!      do ii=1, nf, 1
!          anf=-2.0d0*af*lamdaf*lf*dcos(ii*pi*lamdaf/2.0d0/lf)**2/pi/
!     +        ((ii*lamdaf)**2-lf**2)
!          bnf=-af*lamdaf*lf*dsin(ii*pi*lamdaf/lf)/pi/
!     +        ((ii*lamdaf)**2-lf**2)
!          bps=bps+anf*dcos(ii*2.0d0*pi*xc/lf)+
!     +          bnf*dsin(ii*2.0d0*pi*xc/lf)
!          bps2d=bps2d-(anf*dcos(ii*2.0d0*pi*xc/lf)+
!     +          bnf*dsin(ii*2.0d0*pi*xc/lf))*(ii*2.0d0*pi/lf)**2
!          bps1d=bps1d+(bnf*dcos(ii*2.0d0*pi*xc/lf)-
!     +          anf*dsin(ii*2.0d0*pi*xc/lf))*(ii*2.0d0*pi/lf)
!      enddo

!      Road roughness, for sine wave form
!    two=2.0d0
!    pi=3.141592654d0
!    rough=0.001d0
!    wvln=0.057d0
!    phag=0.0d0
!    bps=rough*dsin(two*pi*dabs(xc)/wvln+phag)
!    bps1d=rough*(two*pi/wvln)*dcos(two*pi*dabs(xc)/wvln+phag)
!    bps2d=-rough*((two*pi/wvln)**2)*dsin(two*pi*dabs(xc)/wvln+phag)

!     Road roughness, for random roughness using Au's method
!      amp=0.0d-8
!      wn1=0.0233d0
!      wn2=0.131d0
!     wnu=6.28d0
!      wnl=0.01d0
      
!      call fixrandom (fq)
!      fq=fq*6.2831853d0
      
!      bps=0.0d0
!      bps1d=0.0d0
!      bps2d=0.0d0
!      do ii=1, nsp, 1
!          wnstep=(wnu-wnl)/nsp
!          wnk=wnl+(ii-0.5d0)*wnstep
!          psd=amp*wn2*wn2*(wnk**2+wn1**2)/wnk**4/(wnk**2+wn2**2)
!          ak=2.0d0*dsqrt(psd*wnstep)
!          phag=fq(ii)
!          bps=bps+ak*dcos(wnk*dabs(xc)+phag)
!          bps1d=bps1d-ak*wnk*dsin(wnk*dabs(xc)+phag)
!          bps2d=bps2d-ak*wnk*wnk*dcos(wnk*dabs(xc)+phag)
!      enddo
 
        !     Road roughness, for random roughness using Code's method
      ! wnu -- upper limit of the considered wavelength, unit: 1/m
      ! wnl -- lower limit of the considered wavelength, unit: 1/m
      ! nsp -- number of harmonics
      ! Grade A -- amp=16d-6 m^3
      ! Grade B -- amp=64d-6 m^3
      ! Grade C -- amp=256d-6 m^3
      ! Grade D -- amp=1024d-6 m^3
      ! Grade E -- amp=4096d-6 m^3
      ! Grade F -- amp=16384d-6 m^3
      ! Grade G -- amp=65536d-6 m^3
      pi=3.141592654d0
      nsp=500
      amp=16d-6
      wnu=6d0
      wnl=0.02857d0
      wnstep=(wnu-wnl)/nsp
      
      call fixrandom (fq)
      fq=fq*2.0d0*pi
      
      bps=0.0d0
      bps1d=0.0d0
      bps2d=0.0d0
      do ii=1, nsp, 1
          wnk=wnl+(ii-0.5d0)*wnstep
          psd=amp*(wnk/0.1d0)**(-2)
          ak=dsqrt(2.0d0*psd*wnstep)
          phag=fq(ii)
          bps=bps+ak*dcos(2.0d0*pi*wnk*xc+phag)
          bps1d=bps1d-
     +          ak*2.0d0*pi*wnk*dsin(2.0d0*pi*wnk*xc+phag)
          bps2d=bps2d-
     +          ak*dcos(2.0d0*pi*wnk*xc+phag)*((2.0d0*pi*wnk)**2)
       enddo
      
      return
      end subroutine roadrough
************************************************************************      
      subroutine cordtran (mcrd,ndim,nnode,coords,coordjb)
      ! Transform global coordinates to local ones
      ! The global coordinate system is defined as follows
      ! Note: 1. the x' axle of VBI element is arranged along the positive
      !          z axle of the global coordinate system;
      !       2. base of the VBI element of rectangular cross section is parallel
      !          to the x-z plane of the global coordinate, so no geometrical rotation
      !           is cosidered for cross section of the beam, i.e. gamrad=0
      !       3. the details can be found in the book <Programming the Finite Element Method, 4th Edition>
      !          by I.M. Smith and D.V. Griffiths. (subroutine rigid_jointed.f95, P135, for "non-vertical" elements)
      !
      !                   ^y
      !                   |    
      !                   |        
      !                   |
      !                   |  
      !                   |
      !                   ---------------->x
      !                    \
      !                     \
      !                      \
      !                       \
      !                        \ z
      implicit none
      integer mcrd,ndim,nnode
      real*8 coords(mcrd,nnode),coordjb(ndim,nnode),tranjz(3,3)
      real*8 cg,sg,den,XL,YL,ZL,x1,y1,z1,x2,y2,z2,ell,gamrad
      
      integer iter,jter
      
      x1=coords(1,1)
      y1=coords(2,1)
      z1=coords(3,1)
      x2=coords(1,2)
      y2=coords(2,2)
      z2=coords(3,2)
      XL=x2-x1
      YL=y2-y1
      ZL=z2-z1
      ell=DSQRT(XL*XL+YL*YL+ZL*ZL)
      
      gamrad=0.0d0
      cg=dcos(gamrad)
      sg=dsin(gamrad)
      den=ell*DSQRT(XL*XL+ZL*ZL)
      
      if (den .ne. 0.0d0) then
          tranjz(1,1)=XL/ell
          tranjz(1,2)=YL/ell
          tranjz(1,3)=ZL/ell
          tranjz(2,1)=(-XL*YL*cg-ell*ZL*sg)/den
          tranjz(2,2)=den*cg/(ell*ell)
          tranjz(2,3)=(-YL*ZL*cg+ell*XL*sg)/den
          tranjz(3,1)=(XL*YL*sg-ell*ZL*cg)/den
          tranjz(3,2)=-den*sg/(ell*ell)
          tranjz(3,3)=(YL*ZL*sg+ell*XL*cg)/den
      else
          tranjz(1,1)=0.0d0
          tranjz(1,3)=0.0d0
          tranjz(2,2)=0.0d0
          tranjz(3,2)=0.0d0
          tranjz(1,2)=1.0d0
          tranjz(2,1)=-cg
          tranjz(3,3)=cg
          tranjz(2,3)=sg
          tranjz(3,1)=sg
      endif
      
      
      do iter=1, nnode, 1
          do jter=1, ndim, 1
              coordjb(jter,iter)=tranjz(jter,1)*coords(1,iter)+ 
     +                           tranjz(jter,2)*coords(2,iter)+
     +                           tranjz(jter,3)*coords(3,iter)
          enddo
      enddo
      
      return
      end subroutine cordtran
      
************************************************************************      
      subroutine eledoftran (mcrd,nnode,coords,NDOFEL,Uall,Vel,
     *                       Accn,ujb,vjb,ajb)
      ! Transform displacement, velocity and acceleratio responses in 
      ! global coordinates to local ones.
      ! The global coordinate system is defined as follows
      ! Note: 1. the x' axle of VBI element is arranged along the positive
      !          z axle of the global coordinate system;
      !       2. base of the VBI element of rectangular cross section is parallel
      !          to the x-z plane of the global coordinate, so no geometrical rotation
      !           is cosidered for cross section of the beam, i.e. gamrad=0
      !       3. the details can be found in the book <Programming the Finite Element Method, 4th Edition>
      !          by I.M. Smith and D.V. Griffiths. (subroutine rigid_jointed.f95, P135, for "non-vertical" elements)
      !
      !                   ^y
      !                   |    
      !                   |        
      !                   |
      !                   |  
      !                   |
      !                   ---------------->x
      !                    \
      !                     \
      !                      \
      !                       \
      !                        \ z
      implicit none
      integer ndofel,mcrd,nnode
      real*8 Uall(ndofel),Vel(ndofel),Accn(ndofel),coords(mcrd,nnode)
      real*8 ujb(ndofel),vjb(ndofel),ajb(ndofel)
      real*8 trandof(3,3),t(12,12)
      real*8 cg,sg,den,XL,YL,ZL,x1,y1,z1,x2,y2,z2,ell,gamrad
      integer i,j,k
      real*8 x
      
      x1=coords(1,1)
      y1=coords(2,1)
      z1=coords(3,1)
      x2=coords(1,2)
      y2=coords(2,2)
      z2=coords(3,2)
      XL=x2-x1
      YL=y2-y1
      ZL=z2-z1
      ell=DSQRT(XL*XL+YL*YL+ZL*ZL)
      
      gamrad=0.0d0
      cg=dcos(gamrad)
      sg=dsin(gamrad)
      den=ell*DSQRT(XL*XL+ZL*ZL)
      
      if (den .ne. 0.0d0) then
          trandof(1,1)=XL/ell
          trandof(1,2)=YL/ell
          trandof(1,3)=ZL/ell
          trandof(2,1)=(-XL*YL*cg-ell*ZL*sg)/den
          trandof(2,2)=den*cg/(ell*ell)
          trandof(2,3)=(-YL*ZL*cg+ell*XL*sg)/den
          trandof(3,1)=(XL*YL*sg-ell*ZL*cg)/den
          trandof(3,2)=-den*sg/(ell*ell)
          trandof(3,3)=(YL*ZL*sg+ell*XL*cg)/den
      else
          trandof(1,1)=0.0d0
          trandof(1,3)=0.0d0
          trandof(2,2)=0.0d0
          trandof(3,2)=0.0d0
          trandof(1,2)=1.0d0
          trandof(2,1)=-cg
          trandof(3,3)=cg
          trandof(2,3)=sg
          trandof(3,1)=sg
      endif
      
      t=0.0d0
      do i=1,3
          do j=1,3
              x=trandof(i,j)
              do k=0,9,3
                  t(i+k,j+k)=x
              enddo
          enddo
      enddo
      
      ujb=matmul(t,Uall)
      vjb=matmul(t,Vel)
      ajb=matmul(t,Accn)
      
      return
      end subroutine eledoftran
************************************************************************      
      
      subroutine amarhstran (mcrd,nnode,coords,MLVARX,NDOFEL,AMATRX,rhs)
      ! Transform "amatrx" and "rhs" that are formed in local coordinates
      ! into global coordinates.
      ! The global coordinate system is defined as follows
      ! Note: 1. the x' axle of VBI element is arranged along the positive
      !          z axle of the global coordinate system;
      !       2. base of the VBI element of rectangular cross section is parallel
      !          to the x-z plane of the global coordinate, so no geometrical rotation
      !           is cosidered for cross section of the beam, i.e. gamrad=0
      !       3. the details can be found in the book <Programming the Finite Element Method, 4th Edition>
      !          by I.M. Smith and D.V. Griffiths. (subroutine rigid_jointed.f95, P135, for "non-vertical" elements)
      !
      !                   ^y
      !                   |    
      !                   |        
      !                   |
      !                   |  
      !                   |
      !                   ---------------->x
      !                    \
      !                     \
      !                      \
      !                       \
      !                        \ z
      implicit none
      integer ndofel,mlvarx,mcrd,nnode
      real*8 rhs(mlvarx,*),amatrx(ndofel,ndofel),coords(mcrd,nnode)
      real*8 trandof(3,3),t(12,12),tt(12,12)
      real*8 cg,sg,den,XL,YL,ZL,x1,y1,z1,x2,y2,z2,ell,gamrad
      integer i,j,k
      real*8 x,sum,cc(12,12)
      
      x1=coords(1,1)
      y1=coords(2,1)
      z1=coords(3,1)
      x2=coords(1,2)
      y2=coords(2,2)
      z2=coords(3,2)
      XL=x2-x1
      YL=y2-y1
      ZL=z2-z1
      ell=DSQRT(XL*XL+YL*YL+ZL*ZL)
      
      gamrad=0.0d0
      cg=dcos(gamrad)
      sg=dsin(gamrad)
      den=ell*DSQRT(XL*XL+ZL*ZL)
      
      if (den .ne. 0.0d0) then
          trandof(1,1)=XL/ell
          trandof(1,2)=YL/ell
          trandof(1,3)=ZL/ell
          trandof(2,1)=(-XL*YL*cg-ell*ZL*sg)/den
          trandof(2,2)=den*cg/(ell*ell)
          trandof(2,3)=(-YL*ZL*cg+ell*XL*sg)/den
          trandof(3,1)=(XL*YL*sg-ell*ZL*cg)/den
          trandof(3,2)=-den*sg/(ell*ell)
          trandof(3,3)=(YL*ZL*sg+ell*XL*cg)/den
      else
          trandof(1,1)=0.0d0
          trandof(1,3)=0.0d0
          trandof(2,2)=0.0d0
          trandof(3,2)=0.0d0
          trandof(1,2)=1.0d0
          trandof(2,1)=-cg
          trandof(3,3)=cg
          trandof(2,3)=sg
          trandof(3,1)=sg
      endif
      
      
      !!** Output for check
      !write(80,*) 'trandof(1,1)=',trandof(1,1)
      !write(80,*) 'trandof(1,2)=',trandof(1,2)
      !write(80,*) 'trandof(1,3)=',trandof(1,3)
      !write(80,*) 'trandof(2,1)=',trandof(2,1)
      !write(80,*) 'trandof(2,2)=',trandof(2,2)
      !write(80,*) 'trandof(2,3)=',trandof(2,3)
      !write(80,*) 'trandof(3,1)=',trandof(3,1)
      !write(80,*) 'trandof(3,2)=',trandof(3,2)
      !write(80,*) 'trandof(3,3)=',trandof(3,3)
      
      t=0.0d0
      tt=0.0d0
      do i=1,3
          do j=1,3
              x=trandof(i,j)
              do k=0,9,3
                  t(i+k,j+k)=x
                  tt(j+k,i+k)=x
              enddo
          enddo
      enddo
      
      !!** Output for check
      !write(80,*) 'amatrx0(1,1)=',amatrx(1,1)
      !write(80,*) 'amatrx0(2,2)=',amatrx(2,2)
      !write(80,*) 'amatrx0(3,3)=',amatrx(3,3)
      !write(80,*) 'amatrx0(4,4)=',amatrx(4,4)
      !write(80,*) 'amatrx0(5,5)=',amatrx(5,5)
      !write(80,*) 'amatrx0(6,6)=',amatrx(6,6)
      !write(80,*) 'amatrx0(7,7)=',amatrx(7,7)
      !write(80,*) 'amatrx0(8,8)=',amatrx(8,8)
      !write(80,*) 'amatrx0(9,9)=',amatrx(9,9)
      !write(80,*) 'amatrx0(10,10)=',amatrx(10,10)
      !write(80,*) 'amatrx0(11,11)=',amatrx(11,11)
      !write(80,*) 'amatrx0(12,12)=',amatrx(12,12)
      
      
      do i=1,ndofel,1
          do j=1,ndofel,1
              sum=0.0d0
              do k=1,ndofel,1
                  sum=sum+amatrx(i,k)*t(k,j)
              enddo
              cc(i,j)=sum
          enddo
      enddo
      do i=1,ndofel,1
          do j=1,ndofel,1
              sum=0.0d0
              do k=1,ndofel,1
                  sum=sum+tt(i,k)*cc(k,j)
              enddo
              amatrx(i,j)=sum
          enddo
      enddo
      
          
      !amatrx=matmul(tt,matmul(amatrx,t))
      rhs(1:ndofel,1)=matmul(tt,rhs(1:ndofel,1))
      
      !!** Output for check
      !write(80,*) 'amatrx(1,1)=',amatrx(1,1)
      !write(80,*) 'amatrx(2,2)=',amatrx(2,2)
      !write(80,*) 'amatrx(3,3)=',amatrx(3,3)
      !write(80,*) 'amatrx(4,4)=',amatrx(4,4)
      !write(80,*) 'amatrx(5,5)=',amatrx(5,5)
      !write(80,*) 'amatrx(6,6)=',amatrx(6,6)
      !write(80,*) 'amatrx(7,7)=',amatrx(7,7)
      !write(80,*) 'amatrx(8,8)=',amatrx(8,8)
      !write(80,*) 'amatrx(9,9)=',amatrx(9,9)
      !write(80,*) 'amatrx(10,10)=',amatrx(10,10)
      !write(80,*) 'amatrx(11,11)=',amatrx(11,11)
      !write(80,*) 'amatrx(12,12)=',amatrx(12,12)
      
      return
      end subroutine amarhstran
************************************************************************      
      subroutine fixrandom (fq)
      
      implicit none
      real*8 fq(500), ran(500)
      
      ! input the random number in the range 0~1 for the random irregularities
      ! so the random number will not be generated per calling the VBI element
      ! i.e. no matter how many calls to the VBI subroutine, the same group of 
      ! random numbers will be used.
      data ran /
     *     0.3920868D-06,
     *     0.2548044D-01,
     *     0.3525162D+00,
     *     0.6669145D+00,
     *     0.9630555D+00,
     *     0.8382882D+00,
     *     0.3353550D+00,
     *     0.9153272D+00,
     *     0.7958637D+00,
     *     0.8326931D+00,
     *     0.3450427D+00,
     *     0.8711839D+00,
     *     0.8991836D-01,
     *     0.8882838D+00,
     *     0.7009789D+00,
     *     0.7345526D+00,
     *     0.3001758D+00,
     *     0.4971772D-01,
     *     0.9081894D+00,
     *     0.9765860D-01,
     *     0.4031338D-01,
     *     0.8502479D-01,
     *     0.5588210D+00,
     *     0.9264517D+00,
     *     0.7564077D-01,
     *     0.9117896D+00,
     *     0.9182218D-01,
     *     0.6377664D+00,
     *     0.8522921D+00,
     *     0.1210777D+00,
     *     0.9945745D+00,
     *     0.7782503D+00,
     *     0.1966502D-01,
     *     0.1698637D+00,
     *     0.9947433D+00,
     *     0.7536925D+00,
     *     0.2179738D+00,
     *     0.6592926D+00,
     *     0.5390184D+00,
     *     0.4800634D+00,
     *     0.8392264D+00,
     *     0.2542021D-01,
     *     0.9893829D+00,
     *     0.5857707D+00,
     *     0.7670679D+00,
     *     0.4448169D+00,
     *     0.7095051D+00,
     *     0.7945261D+00,
     *     0.4387014D+00,
     *     0.2087533D+00,
     *     0.7311122D+00,
     *     0.6657310D+00,
     *     0.7220742D+00,
     *     0.2114710D+00,
     *     0.3082515D+00,
     *     0.2629889D+00,
     *     0.7522959D+00,
     *     0.5497946D+00,
     *     0.6367239D+00,
     *     0.6310104D+00,
     *     0.4947841D+00,
     *     0.4706951D+00,
     *     0.1225150D+00,
     *     0.5348967D+00,
     *     0.1941571D+00,
     *     0.9679053D+00,
     *     0.5892302D+00,
     *     0.1762332D-01,
     *     0.3143427D+00,
     *     0.5533030D+00,
     *     0.6762834D+00,
     *     0.1688328D-01,
     *     0.2656559D+00,
     *     0.4608800D+00,
     *     0.8374500D+00,
     *     0.2289531D+00,
     *     0.6235155D+00,
     *     0.4545669D-01,
     *     0.5232225D+00,
     *     0.8096840D+00,
     *     0.3317850D+00,
     *     0.8919281D+00,
     *     0.7028152D+00,
     *     0.7460327D+00,
     *     0.2590721D+00,
     *     0.6712999D+00,
     *     0.6549743D+00,
     *     0.9354450D+00,
     *     0.9712577D+00,
     *     0.5009523D+00,
     *     0.4298008D+00,
     *     0.3171156D+00,
     *     0.6036843D+00,
     *     0.1687085D+00,
     *     0.9264558D+00,
     *     0.5355654D+00,
     *     0.5377909D+00,
     *     0.5648708D+00,
     *     0.3697167D+00,
     *     0.4880914D+00,
     *     0.1926283D+00,
     *     0.6027865D+00,
     *     0.6390631D+00,
     *     0.4777810D+00,
     *     0.7465967D+00,
     *     0.6608131D+00,
     *     0.9748721D+00,
     *     0.8892629D+00,
     *     0.7838956D+00,
     *     0.9777747D+00,
     *     0.6440447D+00,
     *     0.3074740D+00,
     *     0.2376936D+00,
     *     0.2834814D+00,
     *     0.9393794D+00,
     *     0.2201212D+00,
     *     0.6763431D+00,
     *     0.5480947D+00,
     *     0.2565342D+00,
     *     0.1721693D+00,
     *     0.8438470D+00,
     *     0.8668204D+00,
     *     0.4559926D+00,
     *     0.6078690D+00,
     *     0.6300324D+00,
     *     0.1900745D+00,
     *     0.9055328D-01,
     *     0.7067276D+00,
     *     0.2566722D+00,
     *     0.2946954D+00,
     *     0.3789070D+00,
     *     0.1548104D+00,
     *     0.8683998D+00,
     *     0.9624254D+00,
     *     0.1890570D+00,
     *     0.6721458D+00,
     *     0.2014936D+00,
     *     0.9601979D+00,
     *     0.6016344D+00,
     *     0.7366422D+00,
     *     0.9233369D+00,
     *     0.3531015D+00,
     *     0.9779110D+00,
     *     0.7734580D+00,
     *     0.5692677D+00,
     *     0.1820325D+00,
     *     0.6326688D+00,
     *     0.8092992D+00,
     *     0.1547360D+00,
     *     0.5989119D+00,
     *     0.6586968D+00,
     *     0.3769275D+00,
     *     0.8175066D+00,
     *     0.5448293D+00,
     *     0.9822214D+00,
     *     0.9290161D+00,
     *     0.8121325D+00,
     *     0.5405356D-01,
     *     0.2012400D+00,
     *     0.6313092D+00,
     *     0.1049940D+00,
     *     0.1637335D+00,
     *     0.5185676D+00,
     *     0.2091685D+00,
     *     0.2673674D+00,
     *     0.6946034D-01,
     *     0.2035027D-01,
     *     0.8053467D+00,
     *     0.5414063D+00,
     *     0.9410544D+00,
     *     0.7193360D+00,
     *     0.5404551D+00,
     *     0.7751120D+00,
     *     0.4310162D-01,
     *     0.2514670D+00,
     *     0.6148934D+00,
     *     0.3575117D+00,
     *     0.2218520D+00,
     *     0.9110299D+00,
     *     0.2035810D+00,
     *     0.9552382D+00,
     *     0.8658447D+00,
     *     0.6934069D+00,
     *     0.2385547D+00,
     *     0.3787038D+00,
     *     0.9060328D+00,
     *     0.3894488D+00,
     *     0.5431175D+00,
     *     0.7260312D+00,
     *     0.9028746D+00,
     *     0.3304867D+00,
     *     0.1538122D+00,
     *     0.8302961D+00,
     *     0.6181228D+00,
     *     0.7944962D+00,
     *     0.5082856D-01,
     *     0.7390886D+00,
     *     0.6018003D+00,
     *     0.9461981D+00,
     *     0.2080744D+00,
     *     0.8900219D+00,
     *     0.4785482D+00,
     *     0.3391300D+00,
     *     0.8807417D+00,
     *     0.6451143D+00,
     *     0.5037797D+00,
     *     0.6159491D+00,
     *     0.5597744D+00,
     *     0.3151865D+00,
     *     0.8705598D+00,
     *     0.2096077D+00,
     *     0.1319447D+00,
     *     0.2594135D+00,
     *     0.6870495D+00,
     *     0.2406778D+00,
     *     0.8488911D+00,
     *     0.6752855D+00,
     *     0.1584654D+00,
     *     0.3481405D+00,
     *     0.6141816D+00,
     *     0.4168521D+00,
     *     0.7952200D+00,
     *     0.4014066D+00,
     *     0.5446133D+00,
     *     0.3424538D+00,
     *     0.7678871D+00,
     *     0.2028395D+00,
     *     0.7613854D+00,
     *     0.5354565D+00,
     *     0.4542293D+00,
     *     0.1989698D+00,
     *     0.7525686D+00,
     *     0.2929073D+00,
     *     0.2024058D+00,
     *     0.6340665D+00,
     *     0.3297747D+00,
     *     0.7647143D+00,
     *     0.4268674D+00,
     *     0.7279514D-01,
     *     0.6836402D+00,
     *     0.9707414D+00,
     *     0.8120701D+00,
     *     0.1592209D+00,
     *     0.2003273D+00,
     *     0.4423984D+00,
     *     0.8920852D+00,
     *     0.7925396D+00,
     *     0.9336818D+00,
     *     0.4141159D-01,
     *     0.2312816D+00,
     *     0.4631230D+00,
     *     0.3009745D+00,
     *     0.4111296D+00,
     *     0.5694258D+00,
     *     0.8799947D+00,
     *     0.8871261D+00,
     *     0.7011278D+00,
     *     0.2971230D+00,
     *     0.5154175D+00,
     *     0.5698134D+00,
     *     0.4639606D+00,
     *     0.6347641D+00,
     *     0.5891279D+00,
     *     0.6439180D+00,
     *     0.3500535D+00,
     *     0.5568371D+00,
     *     0.7674269D+00,
     *     0.6644922D+00,
     *     0.3787687D-02,
     *     0.4271937D+00,
     *     0.9668566D+00,
     *     0.6739023D-01,
     *     0.7354542D+00,
     *     0.3377645D+00,
     *     0.6664085D-01,
     *     0.8290395D+00,
     *     0.4142372D+00,
     *     0.5472841D-01,
     *     0.9031719D+00,
     *     0.3031390D+00,
     *     0.1188653D+00,
     *     0.4006944D+00,
     *     0.9483061D+00,
     *     0.4081788D+00,
     *     0.5981329D+00,
     *     0.7388975D+00,
     *     0.9779028D+00,
     *     0.2293794D+00,
     *     0.3640443D+00,
     *     0.3077685D+00,
     *     0.9646138D+00,
     *     0.4737787D+00,
     *     0.4109869D+00,
     *     0.2104347D+00,
     *     0.8952092D+00,
     *     0.6122413D-01,
     *     0.8444461D+00,
     *     0.3233251D+00,
     *     0.3491499D+00,
     *     0.2699094D+00,
     *     0.9740744D+00,
     *     0.9438710D+00,
     *     0.8737243D+00,
     *     0.4741079D+00,
     *     0.3971192D+00,
     *     0.7394554D+00,
     *     0.5301331D+00,
     *     0.5660117D+00,
     *     0.1121303D+00,
     *     0.1000897D+00,
     *     0.4565814D+00,
     *     0.6660249D+00,
     *     0.6282755D+00,
     *     0.7793830D-01,
     *     0.8639455D+00,
     *     0.7342449D+00,
     *     0.7932497D+00,
     *     0.2519171D+00,
     *     0.8952739D+00,
     *     0.8444721D-01,
     *     0.7716715D+00,
     *     0.4317269D-01,
     *     0.4592269D+00,
     *     0.3698182D+00,
     *     0.8890253D+00,
     *     0.3008217D+00,
     *     0.2269690D+00,
     *     0.4345605D-01,
     *     0.8684482D+00,
     *     0.7718439D+00,
     *     0.5967982D+00,
     *     0.9237950D+00,
     *     0.6444225D-01,
     *     0.7427279D+00,
     *     0.1496365D+00,
     *     0.5463172D+00,
     *     0.1500548D+00,
     *     0.3404593D-01,
     *     0.3339327D+00,
     *     0.1129837D+00,
     *     0.2827347D+00,
     *     0.4432436D+00,
     *     0.1158527D+00,
     *     0.4076295D+00,
     *     0.1472398D+00,
     *     0.6406723D-01,
     *     0.2088111D+00,
     *     0.8655201D+00,
     *     0.6713118D-01,
     *     0.4870334D+00,
     *     0.9723531D+00,
     *     0.4013780D+00,
     *     0.2038677D+00,
     *     0.9187338D+00,
     *     0.8916079D+00,
     *     0.5614619D+00,
     *     0.4122003D+00,
     *     0.7412490D+00,
     *     0.1033731D+00,
     *     0.9710247D+00,
     *     0.1633505D+00,
     *     0.4069948D+00,
     *     0.4892462D+00,
     *     0.3723035D+00,
     *     0.8242685D+00,
     *     0.2526660D+00,
     *     0.9481259D+00,
     *     0.5118742D+00,
     *     0.7341040D+00,
     *     0.6390466D+00,
     *     0.3198048D-01,
     *     0.1406696D+00,
     *     0.9408092D+00,
     *     0.2026266D+00,
     *     0.8570510D+00,
     *     0.2260187D+00,
     *     0.5926162D+00,
     *     0.8469469D+00,
     *     0.5561238D+00,
     *     0.6949209D+00,
     *     0.2563563D+00,
     *     0.6442098D+00,
     *     0.8861750D+00,
     *     0.9320775D+00,
     *     0.9998591D-01,
     *     0.8226893D+00,
     *     0.8558239D+00,
     *     0.4944522D+00,
     *     0.6925292D+00,
     *     0.5580104D+00,
     *     0.7982388D+00,
     *     0.6718373D+00,
     *     0.4808812D+00,
     *     0.8082469D+00,
     *     0.8128042D+00,
     *     0.2330582D+00,
     *     0.1013568D+00,
     *     0.7082218D+00,
     *     0.5561858D+00,
     *     0.3490360D+00,
     *     0.3548968D+00,
     *     0.2145474D+00,
     *     0.2597517D+00,
     *     0.8657038D+00,
     *     0.2207591D+00,
     *     0.1865117D+00,
     *     0.2278781D+00,
     *     0.3607483D+00,
     *     0.7340579D+00,
     *     0.3245180D+00,
     *     0.6983861D+00,
     *     0.4031267D+00,
     *     0.4409974D+00,
     *     0.3431045D+00,
     *     0.8278457D+00,
     *     0.7623038D+00,
     *     0.6647359D+00,
     *     0.8786333D+00,
     *     0.8445798D+00,
     *     0.9876453D+00,
     *     0.6550830D+00,
     *     0.8503483D+00,
     *     0.9033895D+00,
     *     0.8463545D+00,
     *     0.2066242D+00,
     *     0.6420349D+00,
     *     0.1369968D+00,
     *     0.1926972D+00,
     *     0.9667018D+00,
     *     0.3606348D+00,
     *     0.1613057D+00,
     *     0.1687405D-01,
     *     0.4975743D+00,
     *     0.9714575D+00,
     *     0.6361935D+00,
     *     0.8837938D+00,
     *     0.8467001D+00,
     *     0.6262990D+00,
     *     0.5958858D-01,
     *     0.3279275D+00,
     *     0.3017568D+00,
     *     0.1237910D+00,
     *     0.9063501D+00,
     *     0.2530639D+00,
     *     0.1650879D+00,
     *     0.2774280D+00,
     *     0.9866198D+00,
     *     0.5764256D+00,
     *     0.5527032D+00,
     *     0.5912869D+00,
     *     0.7655716D+00,
     *     0.8500722D+00,
     *     0.6334951D+00,
     *     0.5752297D+00,
     *     0.1413571D+00,
     *     0.8895106D+00,
     *     0.9015764D-01,
     *     0.9629242D+00,
     *     0.3799386D+00,
     *     0.9648831D+00,
     *     0.1915571D+00,
     *     0.6316480D+00,
     *     0.4519085D+00,
     *     0.4577237D+00,
     *     0.9125778D+00,
     *     0.5966362D+00,
     *     0.4528327D-01,
     *     0.9014446D-01,
     *     0.4953701D+00,
     *     0.6052333D+00,
     *     0.7629853D-01,
     *     0.4733437D+00,
     *     0.8312785D+00,
     *     0.1192119D+00,
     *     0.4813845D-01,
     *     0.6997199D+00,
     *     0.9046561D+00,
     *     0.3154788D+00,
     *     0.5884016D-01,
     *     0.9903965D-01,
     *     0.9623596D+00,
     *     0.8004497D-01,
     *     0.3726796D+00,
     *     0.8109766D+00,
     *     0.9025413D+00,
     *     0.4534157D+00,
     *     0.8272570D+00,
     *     0.3208035D+00,
     *     0.2783420D+00,
     *     0.3477339D+00,
     *     0.1832302D+00,
     *     0.3060624D+00,
     *     0.1592160D+00,
     *     0.8545016D+00,
     *     0.8667993D+00,
     *     0.5721873D+00,
     *     0.3721462D+00,
     *     0.1398261D+00,
     *     0.9799260D+00,
     *     0.5168011D+00/
      fq=ran
      
      return
      end subroutine fixrandom