************************************************************************
!
! User element for vehicle bridge interaction basing on Yang's formula
! Yang Y.B., Wu Y.S. A versatile element for analyzing vehicle bridge
! interaction response. Engineering Structures, 2001, 23: 452-469.
! 
! Hossein Azimi, Khaled Galal, Oscar Pekau. A modified numerical VBI
! element for vehicles with constant velocity including road irregularities.
! Engineering Geology, 2011. This paper is refered for including the road
! irregularties into the VBI formulation. May 27, 2019 by SL @ ZJUT

! Both 2D and 3D 2-node E-B beam element are formulated here.
! The Hermite interpolation function is adopted.

! Nodal variables are the displacements and the rotations.
!
! This subroutine is for the following element types
!  > two-dimensional 2 node E-B beam elements with axial effects.
!  > three-dimensional 2 node E-B beam elements with axial effects.
!
! The stiffness, mass and damping matrix of the E-B element can be
! given explicitly. As a result, no numerical integration will be needed.
! However, the transformation should be performed to transform the above
! matrix that are formed in local coordinated to global coordinates.
!
! No mechanical or traction- and pressure-type boundary conditions
! are going to be applied to this VBI element, since they can be 
! given through the dummy elements inherent to ABAQUS. The element contains a 
! suspended vehicle carriage (rigid mass supported by combination of 
! stiffness and damper) and a rigid wheel mass. 
! Moreover, the irregularities between the wheel and the road/rail is considered. 
! However, the contact stiffness is not included, i.e. the wheel is in direct contact
! with the road. In other words, the contact stiffness is infinite, since in the 
! deformation compatibility equation there is no term associating with spring compression,
! i.e. (P(t)-W)/kH, P(t) is dynamic wheel/road force, kH is stiffness of contact spring.
! 

! Li Shi @ ZJUT, April 18, 2019 

***********************************************************************
!
! User element statement in the input file (set ? values as needed):
!
!  2D elements,finished on April 23, 2019, (Unsymmetric solver should be used)
!  *User Element,Nodes=2,Type=U2,Iproperties=1,Properties=14,Coordinates=2,Variables=44
!  1,2,6,USYM
!
!  3D elements,finished on June 23, 2019, (Unsymmetric solver should be used)
!  *User Element,Nodes=2,Type=U3,Iproperties=1,Properties=15,Coordinates=3,Variables=80
!  1,2,3,4,5,6,USYM
!
!
!     State Variables
!     --------------------------------------------------------------
!     G|t and G|t-deltat should be stored as internal variables, as they
!     contain contributions from the beam, wheel and vehicle. 
!     In the input file 'Variables=44' for 2D VBI element
!     In the input file 'Variables=80' for 3D VBI element

!     Material Properties Vector for 2D VBI element
!     --------------------------------------------------------------
!     rhob   = props(1) ! mass density of beam
!     Areab  = props(2) ! section area of beam
!     EIb    = props(3) ! bending rigidity of beam
!     EAb    = Props(4) ! Axial tension stiffness of beam
!     GJb    = Props(5) ! Torsional stiffness of beam
!     vmass  = props(6) ! mass of vehicle carriage (modeled as rigid mass)
!     wmass  = props(7) ! mass of vehicle wheel (modeled as rigid mass)
!     susk   = props(8) ! stiffness of vehicle suspension
!     susc   = props(9) ! damping of vehicle suspension
!     x0     = props(10) ! Initial position (t=0) for the frontmost wheel mass
!     velo   = props(11) ! Speed of vehicle
!     vlen   = props(12) ! Spacing of two neighboring vehicles
!     raydamp_a = props(13) ! Rayleigh damping factor of beam, proportional to mass
!     raydamp_b  = props(14) ! Rayleigh damping factor of beam, proportional to stiffness
!     nc     = jprops(1) ! Number of vehicles
!***********************************************************************
!     Material Properties Vector for 3D VBI element
!     --------------------------------------------------------------
!     Local coordinate system of the E-B beam element is defined as
!     This definition is taken from <Programming the Finite Element Method, 4th Edition>
!     by I.M. Smith and D.V. Griffiths, and this coordinate system does not obey the right-rand rule.
!                   ^y'
!                   |    
!                   |        
!                   |
!                   |  
!                   |
!                   ---------------->x'
!                    \
!                     \
!                      \
!                       \
!                        \ z'
!
!     x', y' and z' are local coordinates associating with this E-B element, where x' denotes the long
!     axis of the element, y' and z' are coordinates defined on cross section of the beam. For beam of 
!     rectangular section, the y'-z' coordinates are defined as follows, i.e. y' and z' are along the
!     height and width of the rectangular cross section.
!                   ^y'
!                   |        
!                   |
!                   |  
!                   |   
!           -----------------
!           |               |
! z'<-------|               |
!           |               |
!           -----------------
!
!
!
!     Suppose 1. the wheel is moving along positive x' axle of the E-B beam element
!             2. the wheel and the suspended vehicle mass is acted along the negative y' axle
!     --------------------------------------------------------------
!     rhob   = props(1) ! mass density of beam
!     Areab  = props(2) ! section area of beam
!     EIby   = props(3) ! bending rigidity of beam about y axle, i.e. bending in x-z plane
!     EIbz   = props(4) ! bending rigidity of beam about z axle, i.e. bending in x-y plane
!     EAbx   = Props(5) ! Axial tension stiffness of beam along x axle
!     GJbx   = Props(6) ! Torsional stiffness of beam about x axle
!     vmass  = props(7) ! mass of vehicle carriage (modeled as rigid mass)
!     wmass  = props(8) ! mass of vehicle wheel (modeled as rigid mass)
!     susk   = props(9) ! stiffness of vehicle suspension
!     susc   = props(10) ! damping of vehicle suspension
!     x0     = props(11) ! Initial position (t=0) for the frontmost wheel mass
!     velo   = props(12) ! Speed of vehicle, a negative value should be assigned if the direction 
!                          of moving is not along the positive local x axle.
!     vlen   = props(13) ! Spacing of two neighboring vehicles, a negative value should be assigned 
!                          if the direction of moving is not along the positive local x axle.
!     raydamp_a = props(14) ! Rayleigh damping factor of beam, proportional to mass
!     raydamp_b  = props(15) ! Rayleigh damping factor of beam, proportional to stiffness
!     nc     = jprops(1) ! Number of vehicles
!***********************************************************************


****************************************************************************

      SUBROUTINE UEL(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +     PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +     TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +     PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +     NJPROP,PERIOD)

*
      IMPLICIT NONE
*
*     VARIABLES DEFINED IN UEL, PASSED BACK TO ABAQUS
*
      REAL(8) :: RHS,AMATRX,SVARS,ENERGY
*
*     VARIABLES PASSED INTO UEL 
*
      REAL(8) :: PROPS,coords,Uall,DUall,Vel,Accn,TIME,
     1  DTIME,PARAMS,ADLMAG,PREDEF,DDLMAG,PNEWDT,PERIOD
      INTEGER :: NDOFEL,NRHS,NSVARS,NPROPS,MCRD,NNODE,JTYPE,KSTEP,KINC,
     1  JELEM,NDLOAD,JDLTYP,NPREDF,LFLAGS,MLVARX,MDLOAD,JPROPS,NJPROP
*
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),coords(MCRD,NNODE),Uall(NDOFEL),
     2 DUall(MLVARX,*),Vel(NDOFEL),Accn(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)

      integer lenJobName,lenOutDir,nDim
      character*256 jobName,outDir,fileName





      !----------------------------------------------------------------
      ! 
      ! Perform initial checks
      !
      !
      ! Open the debug/error message file
      !
      call getJobName(jobName,lenJobName)
      call getOutDir(outDir,lenOutDir)
      fileName = outDir(1:lenOutDir)//'\aaMSGS_'//
     +     jobName(1:lenJobName)//'.dat'
      open(unit=80,file=fileName,status='unknown')


      ! Check the procedure type, this should be an implicit direct
      ! integration procedure,
      !  which are any of the following (11,12)
      !
      if((lflags(1).eq.11).or.(lflags(1).eq.12)) then
         !
         ! all is good
         !
      else
         write(*,*) 'Abaqus does not have the right procedure'
         write(*,*) 'go back and chekc the procedure type'
         write(*,*) 'lflags(1)=',lflags(1)
         write(80,*) 'Abaqus does not have the right procedure'
         write(80,*) 'go back and chekc the procedure type'
         write(80,*) 'lflags(1)=',lflags(1)
         call xit
      endif


      ! Check to see if you are doing a general
      !  step or a linear purturbation step
      !
      if(lflags(4).eq.1) then
         !
         ! lflags(4)=0 -> general step
         ! lflags(4)=1 -> linear purturbation step
         !
         write(*,*) 'Abaqus thinks you are doing'
         write(*,*) 'a linear purturbation step'
         write(80,*) 'Abaqus thinks you are doing'
         write(80,*) 'a linear purturbation step'
         call xit         
      endif


      ! Do nothing if a ``dummy'' step
      !
      if(dtime.eq.0.0) return
      !
      ! Done with initial checks
      !
      !----------------------------------------------------------------



      !----------------------------------------------------------------
      ! 
      ! Call the paricular element to perform the analysis
      !
      if(jtype.eq.2) then
         !
         ! This is a 2-D plane strain analysis, two nodes E-B beam element
         !
         nDim = 2
         call UPE2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +        PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +        TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +        PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +        NJPROP,PERIOD,nDim)
         
      elseif(jtype.eq.3) then
         !
         ! This is a 3D analysis, two nodes E-B beam element
         !
         nDim = 3
         call U3D2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +        PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +        TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +        PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +        NJPROP,PERIOD,nDim)
         !
         !
      else
         !
         ! We have a problem...
         !
         write(*,*) 'Element type not supported, jtype=',jtype
         write(80,*) 'Element type not supported, jtype=',jtype
         call xit
         !
      endif
      !
      ! Done with this element, RHS and AMATRX already returned
      !  as output from the specific element routine called
      !
      !----------------------------------------------------------------


      return
      end subroutine uel

************************************************************************

      subroutine UPE2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +     PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +     TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +     PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +     NJPROP,PERIOD,nDim)

*
      IMPLICIT NONE
*
*     VARIABLES DEFINED IN UEL, PASSED BACK TO ABAQUS
*
      REAL(8) :: RHS,AMATRX,SVARS,ENERGY
*
*     VARIABLES PASSED INTO UEL 
*
      REAL(8) :: PROPS,coords,Uall,DUall,Vel,Accn,TIME,
     1  DTIME,PARAMS,ADLMAG,PREDEF,DDLMAG,PNEWDT,PERIOD
      INTEGER :: NDOFEL,NRHS,NSVARS,NPROPS,MCRD,NNODE,JTYPE,KSTEP,KINC,
     1  JELEM,NDLOAD,JDLTYP,NPREDF,LFLAGS,MLVARX,MDLOAD,JPROPS,NJPROP,
     1  ndim
*
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),coords(MCRD,NNODE),Uall(NDOFEL),
     2 DUall(MLVARX,*),Vel(NDOFEL),Accn(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)

      real*8 vmass,wmass,susk,susc,rhob,areab,Eib,raydamp_a,raydamp_b,
     +       x0,velo,vlen,Eab,GJb,eln,x1x,x2x,wgh,
     +       xn,xc,xmax,xmin,stf(ndofel,ndofel),xksi,shph(ndofel,1),
     +       phuu,phwu,mslp(ndofel,ndofel),a0,a1,a2,a3,a4,a5,a6,a7,
     +       alpha,beta,gama,rdamp(ndofel,ndofel),ydx(ndofel),
     +       gcrt(ndofel),glst(ndofel),fxt(ndofel),cntf,wdisp,wvel,wacc,
     +       vinc,vfc,velt,qucrt,qwcrt,qulst,qwlst,qcrt(ndofel),
     +       qlst(ndofel)
      real*8 shcn(1,ndofel), shcn1d(1,ndofel), 
     *       shcn2d(1,ndofel), bps, bps1d, bps2d

      integer*4 nc,i,iflag,j,iter
      
      real*8 zero,one,two,half,Pi,three,third,gacc
      parameter(zero=0.d0,one=1.d0,two=2.d0,half=0.5d0,Pi=3.141592653d0,
     +     three=3.d0,third=1.d0/3.d0,gacc=-9.81d0)
      
      ! Obtain material properties
      rhob=props(1)
      areab=props(2)
      Eib=props(3)
      Eab=props(4)
      GJb=props(5)
      vmass=props(6)
      wmass=props(7)
      susk=props(8)
      susc=props(9)
      x0=props(10)
      velo=props(11)
      vlen=props(12)
      raydamp_a=props(13)
      raydamp_b=props(14)
      nc=jprops(1)
      
      ! Obtain time integration parameters, used in condensing the vehicle
      ! dofs into those of the wheel basing on the Newmark scheme
      alpha=params(1)
      beta=params(2)
      gama=params(3)
      a0=one/beta/dtime/dtime
      a1=one/beta/dtime
      a2=one/two/beta-one
      a3=dtime*(one-gama)
      a4=gama*dtime
      a5=a4*a0
      a6=a4*a1-one
      a7=a4*a2-a3

      ! Suppose the vehicle is moving along x axle
      ! suppose spacing between neighboring vehicles are larger than
      ! the length of beam element, so that a given beam element will
      ! be loaded by ONLY one vehicle.
      x1x=dabs(coords(1,1))
      x2x=dabs(coords(1,2))
      xmax=max(x1x,x2x)
      xmin=min(x1x,x2x)
      eln=xmax-xmin
      if (eln .le. 0.0d0) then
          write(80,*) 'Zero element length, pls check ...'
          call xit
      endif
      iflag=0
      iter=0
      do i=1,nc,1
          xn=x0-(i-1)*vlen
          xc=xn+velo*time(2)
          if (abs(xc).le.xmax .and. abs(xc).ge.xmin) then
              ! This conventional beam element will be transformed
              ! into VBI element
              iter=i
              iflag=1
              xksi=(abs(xc)-xmin)/(xmax-xmin)
              !write(80,*) xc,xmin,xksi
              exit
          endif
      enddo
      
      
              
      ! Form stiffness matrix, upper half
      stf=zero
      stf(1,1)=Eab/eln
      stf(1,4)=-Eab/eln
      stf(2,2)=12.0d0*Eib/(eln**3)
      stf(2,3)=6.0d0*Eib/(eln**2)
      stf(2,5)=-12.0d0*Eib/(eln**3)
      stf(2,6)=6.0d0*Eib/(eln**2)
      stf(3,3)=4.0d0*Eib/eln
      stf(3,5)=-6.0d0*Eib/(eln**2)
      stf(3,6)=2.0d0*Eib/eln
      stf(4,4)=Eab/eln
      stf(5,5)=12.0d0*Eib/(eln**3)
      stf(5,6)=-6.0d0*Eib/(eln**2)
      stf(6,6)=4.0d0*Eib/eln
      
      ! Form mass matrix, lumped mass is used, upper half
      mslp=zero
      mslp(1,1)=rhob*areab*eln/two
      mslp(2,2)=rhob*areab*eln/two
      mslp(4,4)=rhob*areab*eln/two
      mslp(5,5)=rhob*areab*eln/two
      
      ! Form mass matrix, consist mass is used, upper half
      !wgh=rhob*areab*eln
      !mslp=zero
      !mslp(1,1)=third*wgh
      !mslp(1,4)=third*half*wgh
      !mslp(2,2)=156.0d0*wgh/420.0d0
      !mslp(2,3)=-22.0d0*eln*wgh/420.0d0
      !mslp(2,5)=54.0d0*wgh/420.0d0
      !mslp(2,6)=13.0d0*eln*wgh/420.0d0
      !mslp(3,3)=4.0d0*(eln**2)*wgh/420.0d0
      !mslp(3,5)=-13.0d0*eln*wgh/420.0d0
      !mslp(3,6)=-three*(eln**2)*wgh/420.0d0
      !mslp(4,4)=third*wgh
      !mslp(5,5)=156.0d0*wgh/420.0d0
      !mslp(5,6)=22.0d0*eln*wgh/420.0d0
      !mslp(6,6)=4.0d0*(eln**2)*wgh/420.0d0
      
      ! Form stiffness and mass matrix, lower half, use symmetry
      do i=1,6,1
          do j=i+1,6,1
              stf(j,i)=stf(i,j)
              mslp(j,i)=mslp(i,j)
          enddo
      enddo
              
      
      ! Form damp matrix, Rayleigh damping is assumed.
      rdamp=raydamp_a*mslp+raydamp_b*stf
      
      ! Hermite interpolation for determing contribution
      ! of moving vehicle on beam stiffness matrix
      if (iflag .eq. 1) then
          
          ! Hermite interpolation func at current local coord xksi
          shph=zero
          shph(2,1)=one-three*xksi**2+two*xksi**3
          shph(5,1)=three*xksi**2-two*xksi**3
          
          ! Consistent interpolation
          shcn=zero
          shcn(1,2)=shph(2,1)
          shcn(1,5)=shph(5,1)
          shcn(1,3)=xksi*(one-two*xksi+xksi**2)*eln
          shcn(1,6)=xksi*(xksi**2-xksi)*eln
          
          ! First order derivative of shcn with respect to xksi
          shcn1d=zero
          shcn1d(1,2)=two*three*xksi*(xksi-one)
          shcn1d(1,5)=two*three*xksi*(one-xksi)
          shcn1d(1,3)=(one-two*two*xksi+three*xksi**2)*eln
          shcn1d(1,6)=(three*xksi**2-two*xksi)*eln
          
          ! Second order derivative of shcn with respect to xksi
          shcn2d=zero
          shcn2d(1,2)=two*three*(two*xksi-one)
          shcn2d(1,5)=two*three*(one-two*xksi)
          shcn2d(1,3)=(three*xksi-two)*two*eln
          shcn2d(1,6)=(three*xksi-one)*two*eln
          
          
          ! Update stiffness and damping matrix for VBI element
          phuu=a0*vmass+a5*susc+susk
          phwu=-a5*susc-susk
          
          stf=stf+(one+phwu/phuu)*susk*matmul(shph,shcn)
          stf=stf+(wmass*(velo/eln)**2)*matmul(shph,shcn2d)
          stf=stf+(one+phwu/phuu)*susc*(velo/eln)*matmul(shph,shcn1d)

          rdamp=rdamp+(one+phwu/phuu)*susc*matmul(shph,shcn)
          rdamp=rdamp+(two*wmass*velo/eln)*matmul(shph,shcn1d)
          
          ! Update mass matrix for VBI element
          mslp=mslp+wmass*matmul(shph,shcn)

      endif
      ! Forming AMATRX
      AMATRX=a0*mslp+(one+alpha)*a5*rdamp+(one+alpha)*stf
          
      ! SVARS(NSVARS) contains contribution of beam disp, beam velocity
      ! external force on beam, and contribution of moving vehicle only if
      ! the current beamt element is a VBI element, at time t+deltat and t
      ! for internal variables at time t
      ! i.e. SVARS(1:6)=YDX|t=-C*Vel-K*Uall,svars(7:12)=FXT|t;svars(13:18)=q|t;
      !        svars(19:21)={du,duvel,duacc}|t; svars(22)=flag|t;
      ! for internal variables at time t-deltat
      ! i.e. SVARS(22+1:22+6)=YDX|t-deltat,svars(22+7:22+12)=FXT|t-deltat; &
      !      svars(22+13:22+18)=q|t-deltat,svars(22+19:22+21)={du,duvel,duacc}|t-deltat; &
      !      svars(22+22)=flag|t-deltat
      ! i.e. SVARS(NSVARS=44)
      ! G|t+deltat=-C*Vel|t+deltat-K*Uall|t+deltta+Fxt|t+deltat-svars(13:18)
      ! G|t=svars(1:6)+svars(7:12)-svars(35:40)
      ! RHS|t+deltat=-M*Acc|t+deltat+(1+alpha)*G|t+deltat-alpha*G|t
      
      gcrt=zero
      glst=zero
      fxt=zero
      ydx=zero
      
      ydx=-matmul(rdamp,Vel)-matmul(stf,Uall)
      glst=svars(1:6)
      gcrt=ydx
      if (iflag .eq. 1) then
          
          !xc - coordinate of the wheel at global coord sys
          ! bps: amplitude of road roughness at xc
          ! bps1d: the first derivative of road roughness at xc
          ! bps2d: the second derivative of road roughness at xc
          xc=x0-(iter-1)*vlen+velo*time(2)
          call roadrough (xc,bps,bps1d,bps2d)
          
          do i=1,ndofel,1
              fxt(i)=shph(i,1)*(vmass+wmass)*gacc
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*susk*bps
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*susc*bps1d*velo
              fxt(i)=fxt(i)-shph(i,1)*wmass*bps2d*velo*velo
          enddo
          gcrt=gcrt+fxt

          if (int(svars(22)) .eq. 1) then
              gcrt=gcrt-svars(13:18)
              glst=glst+svars(7:12)
          endif
          
          if (int(svars(44)) .eq. 1) then
              glst=glst-svars(35:40)
          endif
          
          svars(29:44)=svars(7:22)
          svars(7:12)=fxt
          
          ! Obtain disp, vel and acc at contact point for t+deltat
          wdisp=zero
          wvel=zero
          wacc=zero
          do i=1,ndofel,1
              wdisp=wdisp+shcn(1,i)*Uall(i)
              wvel=wvel+shcn(1,i)*Vel(i)+velo*shcn1d(1,i)*Uall(i)/eln
              wacc=wacc+shcn(1,i)*Accn(i)+
     *                  two*velo*shcn1d(1,i)*Vel(i)/eln+
     *                  shcn2d(1,i)*Uall(i)*(velo/eln)**2
          enddo
          ! Include contribution of road roughness into wheel response
          wdisp=wdisp+bps
          wvel=wvel+velo*bps1d
          wacc=wacc+velo*velo*bps2d
          
          cntf=-susc*wvel-susk*wdisp
          vfc=vmass*(a1*svars(20)+a2*svars(21))+ 
     +        susc*(a6*svars(20)+a7*svars(21))-susk*svars(19)
          vinc=(vfc-cntf)/phuu
          svars(19)=svars(19)+vinc
          velt=svars(20)
          svars(20)=a5*vinc-a6*velt-a7*svars(21)
          svars(21)=a0*vinc-a1*velt-a2*svars(21)
          svars(22)=real(iflag)
          
          qucrt=vmass*(a1*svars(20)+a2*svars(21))+
     +          susc*(a6*svars(20)+a7*svars(21))-susk*svars(19)
          qwcrt=-susc*(a6*svars(20)+a7*svars(21))+susk*svars(19)
          do i=1,ndofel,1
              qcrt(i)=shph(i,1)*(phwu/phuu*qucrt-qwcrt)
          enddo
          svars(13:18)=qcrt
          
      endif
      rhs(1:ndofel,1)=-matmul(mslp,Accn)+(one+alpha)*gcrt-alpha*glst
      
      ! Update internal variable at the end of current increment
      svars(23:28)=svars(1:6)
      svars(1:6)=ydx
      
      ! Update energy, energy(1): kinematic energy, 
      ! energy(2): elastic strain energy
      energy(1)=zero
      do i=1,2,1
          energy(1)=energy(1)+half*Vel(i)*(half*rhob*areab*eln)*Vel(i)
      enddo
      do i=4,5,1
          energy(1)=energy(1)+half*Vel(i)*(half*rhob*areab*eln)*Vel(i)
      enddo
      energy(2)=zero
      energy(2)=energy(2)+half*Eab*(Uall(1)-Uall(4))**2/eln
      energy(2)=energy(2)+half*Eib*12.0d0*(Uall(2)-Uall(5))**2/eln**3

      return
      end subroutine UPE2

************************************************************************
      subroutine U3D2(RHS,AMATRX,SVARS,ENERGY,NDOFEL,NRHS,NSVARS,
     +        PROPS,NPROPS,coords,MCRD,NNODE,Uall,DUall,Vel,Accn,JTYPE,
     +        TIME,DTIME,KSTEP,KINC,JELEM,PARAMS,NDLOAD,JDLTYP,ADLMAG,
     +        PREDEF,NPREDF,LFLAGS,MLVARX,DDLMAG,MDLOAD,PNEWDT,JPROPS,
     +        NJPROP,PERIOD,nDim)  
*
      IMPLICIT NONE
*
*     VARIABLES DEFINED IN UEL, PASSED BACK TO ABAQUS
*
      REAL(8) :: RHS,AMATRX,SVARS,ENERGY
*
*     VARIABLES PASSED INTO UEL 
*
      REAL(8) :: PROPS,coords,Uall,DUall,Vel,Accn,TIME,
     1  DTIME,PARAMS,ADLMAG,PREDEF,DDLMAG,PNEWDT,PERIOD
      INTEGER :: NDOFEL,NRHS,NSVARS,NPROPS,MCRD,NNODE,JTYPE,KSTEP,KINC,
     1  JELEM,NDLOAD,JDLTYP,NPREDF,LFLAGS,MLVARX,MDLOAD,JPROPS,NJPROP,
     1  ndim
*
      DIMENSION RHS(MLVARX,*),AMATRX(NDOFEL,NDOFEL),PROPS(*),
     1 SVARS(*),ENERGY(8),coords(MCRD,NNODE),Uall(NDOFEL),
     2 DUall(MLVARX,*),Vel(NDOFEL),Accn(NDOFEL),TIME(2),PARAMS(*),
     3 JDLTYP(MDLOAD,*),ADLMAG(MDLOAD,*),DDLMAG(MDLOAD,*),
     4 PREDEF(2,NPREDF,NNODE),LFLAGS(*),JPROPS(*)
      
      real*8 vmass,wmass,susk,susc,rhob,areab,Eiby,Eibz,raydamp_a,
     +       raydamp_b,x0,velo,vlen,Eabx,GJbx,eln,x1x,x2x,wgh,
     +       xn,xc,xmax,xmin,stf(ndofel,ndofel),xksi,shph(ndofel,1),
     +       phuu,phwu,mslp(ndofel,ndofel),a0,a1,a2,a3,a4,a5,a6,a7,
     +       alpha,beta,gama,rdamp(ndofel,ndofel),ydx(ndofel),
     +       gcrt(ndofel),glst(ndofel),fxt(ndofel),cntf,wdisp,wvel,wacc,
     +       vinc,vfc,velt,qucrt,qwcrt,qulst,qwlst,qcrt(ndofel),
     +       qlst(ndofel),coordjb(ndim,nnode),ujb(ndofel),vjb(ndofel),
     +       ajb(ndofel)
      real*8 shcn(1,ndofel), shcn1d(1,ndofel), 
     *       shcn2d(1,ndofel), bps, bps1d, bps2d
      real*8 b1,b2,b3,b4,b5,b6,b7,b8

      integer*4 nc,i,iflag,j,iter
      
      real*8 zero,one,two,half,Pi,three,third,gacc
      parameter(zero=0.d0,one=1.d0,two=2.d0,half=0.5d0,Pi=3.141592653d0,
     +     three=3.d0,third=1.d0/3.d0,gacc=-9.81d0)
      
      ! Obtain material properties
      rhob=props(1)
      areab=props(2)
      Eiby=props(3)
      Eibz=props(4)
      Eabx=props(5)
      GJbx=props(6)
      vmass=props(7)
      wmass=props(8)
      susk=props(9)
      susc=props(10)
      x0=props(11)
      velo=props(12)
      vlen=props(13)
      raydamp_a=props(14)
      raydamp_b=props(15)
      nc=jprops(1)
      
      
      
      ! Obtain time integration parameters, used in condensing the vehicle
      ! dofs into those of the wheel basing on the Newmark scheme
      alpha=params(1)
      beta=params(2)
      gama=params(3)
      a0=one/beta/dtime/dtime
      a1=one/beta/dtime
      a2=one/two/beta-one
      a3=dtime*(one-gama)
      a4=gama*dtime
      a5=a4*a0
      a6=a4*a1-one
      a7=a4*a2-a3
      
      ! suppose spacing between neighboring vehicles are larger than
      ! the length of beam element, so that a given beam element will
      ! be loaded by ONLY one vehicle.
      call cordtran(mcrd,ndim,nnode,coords,coordjb)
      x1x=dabs(coordjb(1,1))
      x2x=dabs(coordjb(1,2))
      xmax=max(x1x,x2x)
      xmin=min(x1x,x2x)
      eln=xmax-xmin
      
      if (eln .le. 0.0d0 .or. x2x .lt. x1x) then
          write(80,*) 'Zero element length, pls check ...'
          call xit
      endif
      
      iflag=0
      iter=0
      do i=1,nc,1
          xn=x0-(i-1)*vlen
          xc=xn+velo*time(2)
          if (abs(xc).le.xmax .and. abs(xc).ge.xmin) then
              ! This conventional beam element will be transformed
              ! into VBI element
              iter=i
              iflag=1
              xksi=(abs(xc)-xmin)/(xmax-xmin)
              exit
          endif
      enddo
      
      ! Form stiffness matrix in elemental local coordinates
      ! taken from the book <Programming the Finite Element Method, 4th Edition>
      ! by I.M. Smith and D.V. Griffiths.
      b1=Eabx/eln
      b2=12.0d0*Eibz/(eln**3)
      b3=12.0d0*Eiby/(eln**3)
      b4=6.0d0*Eibz/(eln**2)
      b5=6.0d0*Eiby/(eln**2)
      b6=4.0d0*Eibz/eln
      b7=4.0d0*Eiby/eln
      b8=GJbx/eln
      
      stf=zero
      stf(1,1)=b1
      stf(7,7)=b1
      stf(1,7)=-b1
      stf(7,1)=-b1
      stf(2,2)=b2
      stf(8,8)=b2
      stf(2,8)=-b2
      stf(8,2)=-b2
      stf(3,3)=b3
      stf(9,9)=b3
      stf(3,9)=-b3
      stf(9,3)=-b3
      stf(4,4)=b8
      stf(10,10)=b8
      stf(4,10)=-b8
      stf(10,4)=-b8
      stf(5,5)=b7
      stf(11,11)=b7
      stf(5,11)=half*b7
      stf(11,5)=half*b7
      stf(6,6)=b6
      stf(12,12)=b6
      stf(6,12)=half*b6
      stf(12,6)=half*b6
      stf(2,6)=b4
      stf(6,2)=b4
      stf(2,12)=b4
      stf(12,2)=b4
      stf(6,8)=-b4
      stf(8,6)=-b4
      stf(8,12)=-b4
      stf(12,8)=-b4
      stf(5,9)=b5
      stf(9,5)=b5
      stf(9,11)=b5
      stf(11,9)=b5
      stf(3,5)=-b5
      stf(5,3)=-b5
      stf(3,11)=-b5
      stf(11,3)=-b5
      
      ! Form mass matrix in elemental local coordinates
      ! lumped mass matrix, zero mass warning has been issued by ABAQUS
      wgh=rhob*areab*eln
      mslp=zero
      mslp(1,1)=wgh/two
      mslp(2,2)=wgh/two
      mslp(3,3)=wgh/two
      mslp(4,4)=wgh/two
      mslp(7,7)=wgh/two
      mslp(8,8)=wgh/two
      mslp(9,9)=wgh/two
      mslp(10,10)=wgh/two
      
      ! Form mass matrix in elemental local coordinates
      ! consistent mass matrix
      !wgh=rhob*areab*eln
      !mslp=zero
      !mslp(1,1)=wgh/3.0d0
      !mslp(7,7)=wgh/3.0d0
      !mslp(1,7)=wgh/6.0d0
      !mslp(7,1)=wgh/6.0d0
      !mslp(2,2)=156.0d0*wgh/420.0d0
      !mslp(8,8)=156.0d0*wgh/420.0d0
      !mslp(2,8)=54.0d0*wgh/420.0d0
      !mslp(8,2)=54.0d0*wgh/420.0d0
      !mslp(3,3)=156.0d0*wgh/420.0d0
      !mslp(9,9)=156.0d0*wgh/420.0d0
      !mslp(3,9)=54.0d0*wgh/420.0d0
      !mslp(9,3)=54.0d0*wgh/420.0d0
      !mslp(4,4)=wgh/3.0d0
      !mslp(10,10)=wgh/3.0d0
      !mslp(4,10)=wgh/6.0d0
      !mslp(10,4)=wgh/6.0d0
      !mslp(5,5)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(11,11)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(5,11)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(11,5)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(6,6)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(12,12)=4.0d0*eln*eln*wgh/420.0d0
      !mslp(6,12)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(12,6)=-3.0d0*eln*eln*wgh/420.0d0
      !mslp(2,6)=22.0d0*eln*wgh/420.0d0
      !mslp(6,2)=22.0d0*eln*wgh/420.0d0
      !mslp(2,12)=-13.0d0*eln*wgh/420.0d0
      !mslp(12,2)=-13.0d0*eln*wgh/420.0d0
      !mslp(6,8)=13.0d0*eln*wgh/420.0d0
      !mslp(8,6)=13.0d0*eln*wgh/420.0d0
      !mslp(8,12)=-22.0d0*eln*wgh/420.0d0
      !mslp(12,8)=-22.0d0*eln*wgh/420.0d0
      !mslp(5,9)=13.0d0*eln*wgh/420.0d0
      !mslp(9,5)=13.0d0*eln*wgh/420.0d0
      !mslp(9,11)=-22.0d0*eln*wgh/420.0d0
      !mslp(11,9)=-22.0d0*eln*wgh/420.0d0
      !mslp(3,5)=22.0d0*eln*wgh/420.0d0
      !mslp(5,3)=22.0d0*eln*wgh/420.0d0
      !mslp(3,11)=-13.0d0*eln*wgh/420.0d0
      !mslp(11,3)=-13.0d0*eln*wgh/420.0d0
      
      ! Form damp matrix, Rayleigh damping is assumed.
      rdamp=raydamp_a*mslp+raydamp_b*stf        
              
      ! Hermite interpolation for determing contribution
      ! of moving vehicle on beam stiffness matrix
      if (iflag .eq. 1) then
          
          ! Hermite interpolation func at current local coord xksi
          shph=zero
          shph(2,1)=one-three*xksi**2+two*xksi**3
          shph(8,1)=three*xksi**2-two*xksi**3
          
          ! Consistent interpolation
          shcn=zero
          shcn(1,2)=shph(2,1)
          shcn(1,8)=shph(8,1)
          shcn(1,6)=xksi*(one-two*xksi+xksi**2)*eln
          shcn(1,12)=xksi*(xksi**2-xksi)*eln
          
          ! First order derivative of shcn with respect to xksi
          shcn1d=zero
          shcn1d(1,2)=two*three*xksi*(xksi-one)
          shcn1d(1,8)=two*three*xksi*(one-xksi)
          shcn1d(1,6)=(one-two*two*xksi+three*xksi**2)*eln
          shcn1d(1,12)=(three*xksi**2-two*xksi)*eln
          
          ! Second order derivative of shcn with respect to xksi
          shcn2d=zero
          shcn2d(1,2)=two*three*(two*xksi-one)
          shcn2d(1,8)=two*three*(one-two*xksi)
          shcn2d(1,6)=(three*xksi-two)*two*eln
          shcn2d(1,12)=(three*xksi-one)*two*eln
          
          
          ! Update stiffness and damping matrix for VBI element
          phuu=a0*vmass+a5*susc+susk
          phwu=-a5*susc-susk
          
          stf=stf+(one+phwu/phuu)*susk*matmul(shph,shcn)
          stf=stf+(wmass*(velo/eln)**2)*matmul(shph,shcn2d)
          stf=stf+(one+phwu/phuu)*susc*(velo/eln)*matmul(shph,shcn1d)

          rdamp=rdamp+(one+phwu/phuu)*susc*matmul(shph,shcn)
          rdamp=rdamp+(two*wmass*velo/eln)*matmul(shph,shcn1d)
          
          ! Update mass matrix for VBI element
          mslp=mslp+wmass*matmul(shph,shcn)

      endif
      ! Forming AMATRX
      AMATRX=a0*mslp+(one+alpha)*a5*rdamp+(one+alpha)*stf
      
      
      
      ! SVARS(NSVARS) contains contribution of beam disp, beam velocity
      ! external force on beam, and contribution of moving vehicle only if
      ! the current beamt element is a VBI element, at time t+deltat and t
      ! for internal variables at time t
      ! i.e. SVARS(1:12)=YDX|t=-C*Vel-K*Uall,svars(13:24)=FXT|t;svars(25:36)=q|t;
      !        svars(37:39)={du,duvel,duacc}|t; svars(40)=flag|t;
      ! for internal variables at time t-deltat
      ! i.e. SVARS(40+1:40+12)=YDX|t-deltat,svars(40+13:40+24)=FXT|t-deltat; &
      !      svars(40+25:40+36)=q|t-deltat,svars(40+37:40+39)={du,duvel,duacc}|t-deltat; &
      !      svars(40+40)=flag|t-deltat
      ! i.e. SVARS(NSVARS=80)
      ! G|t+deltat=-C*Vel|t+deltat-K*Uall|t+deltta+Fxt|t+deltat-svars(25:36)
      ! G|t=svars(1:12)+svars(13:24)-svars(65:76)
      ! RHS|t+deltat=-M*Acc|t+deltat+(1+alpha)*G|t+deltat-alpha*G|t
      
      gcrt=zero
      glst=zero
      fxt=zero
      ydx=zero
      
      call eledoftran (mcrd,nnode,coords,NDOFEL,Uall,Vel,
     *                     Accn,ujb,vjb,ajb)
      ydx=-matmul(rdamp,vjb)-matmul(stf,ujb)
      glst=svars(1:12)
      gcrt=ydx
      if (iflag .eq. 1) then
          
          ! xc - coordinate of the wheel at local coord sys
          ! bps: amplitude of road roughness at xc
          ! bps1d: the first derivative of road roughness at xc
          ! bps2d: the second derivative of road roughness at xc
          xc=x0-(iter-1)*vlen+velo*time(2)
          call roadrough (xc,bps,bps1d,bps2d)
          
          do i=1,ndofel,1
              fxt(i)=shph(i,1)*(vmass+wmass)*gacc
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*susk*bps
              fxt(i)=fxt(i)-shph(i,1)*(one+phwu/phuu)*susc*bps1d*velo
              fxt(i)=fxt(i)-shph(i,1)*wmass*bps2d*velo*velo
          enddo
          gcrt=gcrt+fxt

          if (int(svars(40)) .eq. 1) then
              gcrt=gcrt-svars(25:36)
              glst=glst+svars(13:24)
          endif
          
          if (int(svars(80)) .eq. 1) then
              glst=glst-svars(65:76)
          endif
          
          svars(53:80)=svars(13:40)
          svars(13:24)=fxt
          
          ! Obtain disp, vel and acc at contact point for t+deltat
          wdisp=zero
          wvel=zero
          wacc=zero
          call eledoftran (mcrd,nnode,coords,NDOFEL,Uall,Vel,
     *                     Accn,ujb,vjb,ajb)
          do i=1,ndofel,1
              wdisp=wdisp+shcn(1,i)*ujb(i)
              wvel=wvel+shcn(1,i)*vjb(i)+velo*shcn1d(1,i)*ujb(i)/eln
              wacc=wacc+shcn(1,i)*ajb(i)+
     *                  two*velo*shcn1d(1,i)*vjb(i)/eln+
     *                  shcn2d(1,i)*ujb(i)*(velo/eln)**2
          enddo
          ! Include contribution of road roughness into wheel response
          wdisp=wdisp+bps
          wvel=wvel+velo*bps1d
          wacc=wacc+velo*velo*bps2d
          
          cntf=-susc*wvel-susk*wdisp
          vfc=vmass*(a1*svars(38)+a2*svars(39))+ 
     +        susc*(a6*svars(38)+a7*svars(39))-susk*svars(37)
          vinc=(vfc-cntf)/phuu
          svars(37)=svars(37)+vinc
          velt=svars(38)
          svars(38)=a5*vinc-a6*velt-a7*svars(39)
          svars(39)=a0*vinc-a1*velt-a2*svars(39)
          svars(40)=real(iflag)
          
          qucrt=vmass*(a1*svars(38)+a2*svars(39))+
     +          susc*(a6*svars(38)+a7*svars(39))-susk*svars(37)
          qwcrt=-susc*(a6*svars(38)+a7*svars(39))+susk*svars(37)
          do i=1,ndofel,1
              qcrt(i)=shph(i,1)*(phwu/phuu*qucrt-qwcrt)
          enddo
          svars(25:36)=qcrt
          
      endif
      rhs(1:ndofel,1)=-matmul(mslp,ajb)+(one+alpha)*gcrt-alpha*glst
      
      ! Transfer amatrx and rhs that are formed in local coordinates into global coordinates
      call amarhstran (mcrd,nnode,coords,MLVARX,NDOFEL,AMATRX,rhs)
      
      ! Update internal variable at the end of current increment
      svars(41:52)=svars(1:12)
      svars(1:12)=ydx
      
      ! Update energy, energy(1): kinematic energy, 
      ! energy(2): elastic strain energy
      energy(1)=zero
      do i=1,4,1
          energy(1)=energy(1)+half*vjb(i)*(wgh/two)*vjb(i)
      enddo
      do i=7,10,1
          energy(1)=energy(1)+half*vjb(i)*(wgh/two)*vjb(i)
      enddo
      energy(2)=zero
      ! for strain energy contributed by axle deformation along x axle
      energy(2)=energy(2)+half*Eabx*(ujb(1)-ujb(7))**2/eln
      ! for strain energy contributed by torsional deformation about x axle
      energy(2)=energy(2)+half*GJbx*(ujb(4)-ujb(10))**2/eln
      ! for strain energy contributed by bending in x-z plane
      energy(2)=energy(2)+half*Eiby*12.0d0*(ujb(3)-ujb(9))**2/eln**3
      energy(2)=energy(2)+half*Eiby*4.0d0*((eln*ujb(5))**2+
     +          (eln*ujb(11))**2+eln*eln*ujb(5)*ujb(11))/eln**3
      energy(2)=energy(2)+half*Eiby*12.0d0*(eln*ujb(5)+
     +          eln*ujb(11))*(ujb(3)-ujb(9))/eln**3
      ! for strain energy contributed by bending in y-x plane
      energy(2)=energy(2)+half*Eiby*12.0d0*(ujb(2)-ujb(8))**2/eln**3
      energy(2)=energy(2)+half*Eiby*4.0d0*((eln*ujb(6))**2+
     +          (eln*ujb(12))**2+eln*eln*ujb(6)*ujb(12))/eln**3
      energy(2)=energy(2)+half*Eiby*12.0d0*(eln*ujb(6)+
     +          eln*ujb(12))*(ujb(2)-ujb(8))/eln**3
  
      return
      
      end subroutine U3D2
************************************************************************
      subroutine roadrough (xc, bps, bps1d, bps2d)
      
!     ******************************************** 
!     road roughness of sine wave r(x)=rough*Sin(2*Pi*x/wvln+phag) is introduced, May 27, 2019
!      rough: amplitude of road roughness, modeled as a sine wave, m
!      wvln: props(16) ! wavelength of road roughness, modeled as a sine wave, m
!      phag: props(17) ! phase angle of road roughness, modeled as a sine wave, rad
!     ********************************************
!     random roughness is introduced, June 12, 2019, using methods given by Au et al
!     Au et al. Impact study of cable-stayed railway bridges with random rail irregularities,
!               Engineering Structures, 2002, 24: 529-541.
!       Parameters involved in this method are:
!       amp: Parameter of the psd function, unit: m^3
!       wn1: Parameter of the psd function, unit: 1/m
!       wn2: Parameter of the psd function, unit: 1/m
!       wnu: upper limit of the wavenumber, unit: 1/m
!       wnl: lower limit of the wavenumber, unit: 1/m
!       nsp: Number of sampling points, sufficiently large integer 
      IMPLICIT NONE
      real*8 xc, bps, bps1d, bps2d
      real*8 rough, phag, wvln, two, pi
      real*8 amp, wn1, wn2, wnu, wnl, wnk, wnstep, psd, ak
      integer nsp, ii
      real*8,allocatable :: fq(:)

!     Road roughness, for sine wave form
      two=2.0d0
      pi=3.141592654d0
      rough=0.005d0
      wvln=2.0d0
      phag=0.0d0
      bps=rough*dsin(two*pi*dabs(xc)/wvln+phag)
      bps1d=rough*(two*pi/wvln)*dcos(two*pi*dabs(xc)/wvln+phag)
      bps2d=-rough*((two*pi/wvln)**2)*dsin(two*pi*dabs(xc)/wvln+phag)

!!     Road roughness, for random roughness using Au's method
!      amp=2.75d-8
!      wn1=0.0233d0
!      wn2=0.131d0
!      wnu=6.28d0
!      wnl=0.01d0
!      nsp=2000
!      allocate (fq(nsp))
!      do ii=1, nsp, 1
!          call random_number (fq(ii))
!          fq(ii)=fq(ii)*6.2831853d0
!      enddo
!      
!      bps=0.0d0
!      bps1d=0.0d0
!      bps2d=0.0d0
!      do ii=1, nsp, 1
!          wnstep=(wnu-wnl)/nsp
!          wnk=wnl+(ii-0.5d0)*wnstep
!          psd=amp*wn2*wn2*(wnk**2+wn1**2)/wnk**4/(wnk**2+wn2**2)
!          ak=2.0d0*dsqrt(psd*wnstep)
!          phag=fq(ii)
!          bps=bps+ak*dcos(wnk*dabs(xc)+phag)
!          bps1d=bps1d-ak*wnk*dsin(wnk*dabs(xc)+phag)
!          bps2d=bps2d-ak*wnk*wnk*dcos(wnk*dabs(xc)+phag)
!      enddo
!      deallocate (fq)
      
      return
      end subroutine roadrough
************************************************************************      
      subroutine cordtran (mcrd,ndim,nnode,coords,coordjb)
      ! Transform global coordinates to local ones
      ! The global coordinate system is defined as follows
      ! Note: 1. the x' axle of VBI element is arranged along the positive
      !          z axle of the global coordinate system;
      !       2. base of the VBI element of rectangular cross section is parallel
      !          to the x-z plane of the global coordinate, so no geometrical rotation
      !           is cosidered for cross section of the beam, i.e. gamrad=0
      !       3. the details can be found in the book <Programming the Finite Element Method, 4th Edition>
      !          by I.M. Smith and D.V. Griffiths. (subroutine rigid_jointed.f95, P135, for "non-vertical" elements)
      !
      !                   ^y
      !                   |    
      !                   |        
      !                   |
      !                   |  
      !                   |
      !                   ---------------->x
      !                    \
      !                     \
      !                      \
      !                       \
      !                        \ z
      implicit none
      integer mcrd,ndim,nnode
      real*8 coords(mcrd,nnode),coordjb(ndim,nnode),tranjz(3,3)
      real*8 cg,sg,den,XL,YL,ZL,x1,y1,z1,x2,y2,z2,ell,gamrad
      
      integer iter,jter
      
      x1=coords(1,1)
      y1=coords(2,1)
      z1=coords(3,1)
      x2=coords(1,2)
      y2=coords(2,2)
      z2=coords(3,2)
      XL=x2-x1
      YL=y2-y1
      ZL=z2-z1
      ell=DSQRT(XL*XL+YL*YL+ZL*ZL)
      
      gamrad=0.0d0
      cg=dcos(gamrad)
      sg=dsin(gamrad)
      den=ell*DSQRT(XL*XL+ZL*ZL)
      
      if (den .ne. 0.0d0) then
          tranjz(1,1)=XL/ell
          tranjz(1,2)=YL/ell
          tranjz(1,3)=ZL/ell
          tranjz(2,1)=(-XL*YL*cg-ell*ZL*sg)/den
          tranjz(2,2)=den*cg/(ell*ell)
          tranjz(2,3)=(-YL*ZL*cg+ell*XL*sg)/den
          tranjz(3,1)=(XL*YL*sg-ell*ZL*cg)/den
          tranjz(3,2)=-den*sg/(ell*ell)
          tranjz(3,3)=(YL*ZL*sg+ell*XL*cg)/den
      else
          tranjz(1,1)=0.0d0
          tranjz(1,3)=0.0d0
          tranjz(2,2)=0.0d0
          tranjz(3,2)=0.0d0
          tranjz(1,2)=1.0d0
          tranjz(2,1)=-cg
          tranjz(3,3)=cg
          tranjz(2,3)=sg
          tranjz(3,1)=sg
      endif
      
      
      do iter=1, nnode, 1
          do jter=1, ndim, 1
              coordjb(jter,iter)=tranjz(jter,1)*coords(1,iter)+ 
     +                           tranjz(jter,2)*coords(2,iter)+
     +                           tranjz(jter,3)*coords(3,iter)
          enddo
      enddo
      
      return
      end subroutine cordtran
      
************************************************************************      
      subroutine eledoftran (mcrd,nnode,coords,NDOFEL,Uall,Vel,
     *                       Accn,ujb,vjb,ajb)
      ! Transform displacement, velocity and acceleratio responses in 
      ! global coordinates to local ones.
      ! The global coordinate system is defined as follows
      ! Note: 1. the x' axle of VBI element is arranged along the positive
      !          z axle of the global coordinate system;
      !       2. base of the VBI element of rectangular cross section is parallel
      !          to the x-z plane of the global coordinate, so no geometrical rotation
      !           is cosidered for cross section of the beam, i.e. gamrad=0
      !       3. the details can be found in the book <Programming the Finite Element Method, 4th Edition>
      !          by I.M. Smith and D.V. Griffiths. (subroutine rigid_jointed.f95, P135, for "non-vertical" elements)
      !
      !                   ^y
      !                   |    
      !                   |        
      !                   |
      !                   |  
      !                   |
      !                   ---------------->x
      !                    \
      !                     \
      !                      \
      !                       \
      !                        \ z
      implicit none
      integer ndofel,mcrd,nnode
      real*8 Uall(ndofel),Vel(ndofel),Accn(ndofel),coords(mcrd,nnode)
      real*8 ujb(ndofel),vjb(ndofel),ajb(ndofel)
      real*8 trandof(3,3),t(12,12)
      real*8 cg,sg,den,XL,YL,ZL,x1,y1,z1,x2,y2,z2,ell,gamrad
      integer i,j,k
      real*8 x
      
      x1=coords(1,1)
      y1=coords(2,1)
      z1=coords(3,1)
      x2=coords(1,2)
      y2=coords(2,2)
      z2=coords(3,2)
      XL=x2-x1
      YL=y2-y1
      ZL=z2-z1
      ell=DSQRT(XL*XL+YL*YL+ZL*ZL)
      
      gamrad=0.0d0
      cg=dcos(gamrad)
      sg=dsin(gamrad)
      den=ell*DSQRT(XL*XL+ZL*ZL)
      
      if (den .ne. 0.0d0) then
          trandof(1,1)=XL/ell
          trandof(1,2)=YL/ell
          trandof(1,3)=ZL/ell
          trandof(2,1)=(-XL*YL*cg-ell*ZL*sg)/den
          trandof(2,2)=den*cg/(ell*ell)
          trandof(2,3)=(-YL*ZL*cg+ell*XL*sg)/den
          trandof(3,1)=(XL*YL*sg-ell*ZL*cg)/den
          trandof(3,2)=-den*sg/(ell*ell)
          trandof(3,3)=(YL*ZL*sg+ell*XL*cg)/den
      else
          trandof(1,1)=0.0d0
          trandof(1,3)=0.0d0
          trandof(2,2)=0.0d0
          trandof(3,2)=0.0d0
          trandof(1,2)=1.0d0
          trandof(2,1)=-cg
          trandof(3,3)=cg
          trandof(2,3)=sg
          trandof(3,1)=sg
      endif
      
      t=0.0d0
      do i=1,3
          do j=1,3
              x=trandof(i,j)
              do k=0,9,3
                  t(i+k,j+k)=x
              enddo
          enddo
      enddo
      
      ujb=matmul(t,Uall)
      vjb=matmul(t,Vel)
      ajb=matmul(t,Accn)
      
      return
      end subroutine eledoftran
************************************************************************      
      
      subroutine amarhstran (mcrd,nnode,coords,MLVARX,NDOFEL,AMATRX,rhs)
      ! Transform "amatrx" and "rhs" that are formed in local coordinates
      ! into global coordinates.
      ! The global coordinate system is defined as follows
      ! Note: 1. the x' axle of VBI element is arranged along the positive
      !          z axle of the global coordinate system;
      !       2. base of the VBI element of rectangular cross section is parallel
      !          to the x-z plane of the global coordinate, so no geometrical rotation
      !           is cosidered for cross section of the beam, i.e. gamrad=0
      !       3. the details can be found in the book <Programming the Finite Element Method, 4th Edition>
      !          by I.M. Smith and D.V. Griffiths. (subroutine rigid_jointed.f95, P135, for "non-vertical" elements)
      !
      !                   ^y
      !                   |    
      !                   |        
      !                   |
      !                   |  
      !                   |
      !                   ---------------->x
      !                    \
      !                     \
      !                      \
      !                       \
      !                        \ z
      implicit none
      integer ndofel,mlvarx,mcrd,nnode
      real*8 rhs(mlvarx,*),amatrx(ndofel,ndofel),coords(mcrd,nnode)
      real*8 trandof(3,3),t(12,12),tt(12,12)
      real*8 cg,sg,den,XL,YL,ZL,x1,y1,z1,x2,y2,z2,ell,gamrad
      integer i,j,k
      real*8 x,sum,cc(12,12)
      
      x1=coords(1,1)
      y1=coords(2,1)
      z1=coords(3,1)
      x2=coords(1,2)
      y2=coords(2,2)
      z2=coords(3,2)
      XL=x2-x1
      YL=y2-y1
      ZL=z2-z1
      ell=DSQRT(XL*XL+YL*YL+ZL*ZL)
      
      gamrad=0.0d0
      cg=dcos(gamrad)
      sg=dsin(gamrad)
      den=ell*DSQRT(XL*XL+ZL*ZL)
      
      if (den .ne. 0.0d0) then
          trandof(1,1)=XL/ell
          trandof(1,2)=YL/ell
          trandof(1,3)=ZL/ell
          trandof(2,1)=(-XL*YL*cg-ell*ZL*sg)/den
          trandof(2,2)=den*cg/(ell*ell)
          trandof(2,3)=(-YL*ZL*cg+ell*XL*sg)/den
          trandof(3,1)=(XL*YL*sg-ell*ZL*cg)/den
          trandof(3,2)=-den*sg/(ell*ell)
          trandof(3,3)=(YL*ZL*sg+ell*XL*cg)/den
      else
          trandof(1,1)=0.0d0
          trandof(1,3)=0.0d0
          trandof(2,2)=0.0d0
          trandof(3,2)=0.0d0
          trandof(1,2)=1.0d0
          trandof(2,1)=-cg
          trandof(3,3)=cg
          trandof(2,3)=sg
          trandof(3,1)=sg
      endif
      
      
      !!** Output for check
      !write(80,*) 'trandof(1,1)=',trandof(1,1)
      !write(80,*) 'trandof(1,2)=',trandof(1,2)
      !write(80,*) 'trandof(1,3)=',trandof(1,3)
      !write(80,*) 'trandof(2,1)=',trandof(2,1)
      !write(80,*) 'trandof(2,2)=',trandof(2,2)
      !write(80,*) 'trandof(2,3)=',trandof(2,3)
      !write(80,*) 'trandof(3,1)=',trandof(3,1)
      !write(80,*) 'trandof(3,2)=',trandof(3,2)
      !write(80,*) 'trandof(3,3)=',trandof(3,3)
      
      t=0.0d0
      tt=0.0d0
      do i=1,3
          do j=1,3
              x=trandof(i,j)
              do k=0,9,3
                  t(i+k,j+k)=x
                  tt(j+k,i+k)=x
              enddo
          enddo
      enddo
      
      !!** Output for check
      !write(80,*) 'amatrx0(1,1)=',amatrx(1,1)
      !write(80,*) 'amatrx0(2,2)=',amatrx(2,2)
      !write(80,*) 'amatrx0(3,3)=',amatrx(3,3)
      !write(80,*) 'amatrx0(4,4)=',amatrx(4,4)
      !write(80,*) 'amatrx0(5,5)=',amatrx(5,5)
      !write(80,*) 'amatrx0(6,6)=',amatrx(6,6)
      !write(80,*) 'amatrx0(7,7)=',amatrx(7,7)
      !write(80,*) 'amatrx0(8,8)=',amatrx(8,8)
      !write(80,*) 'amatrx0(9,9)=',amatrx(9,9)
      !write(80,*) 'amatrx0(10,10)=',amatrx(10,10)
      !write(80,*) 'amatrx0(11,11)=',amatrx(11,11)
      !write(80,*) 'amatrx0(12,12)=',amatrx(12,12)
      
      
      do i=1,ndofel,1
          do j=1,ndofel,1
              sum=0.0d0
              do k=1,ndofel,1
                  sum=sum+amatrx(i,k)*t(k,j)
              enddo
              cc(i,j)=sum
          enddo
      enddo
      do i=1,ndofel,1
          do j=1,ndofel,1
              sum=0.0d0
              do k=1,ndofel,1
                  sum=sum+tt(i,k)*cc(k,j)
              enddo
              amatrx(i,j)=sum
          enddo
      enddo
      
          
      !amatrx=matmul(tt,matmul(amatrx,t))
      rhs(1:ndofel,1)=matmul(tt,rhs(1:ndofel,1))
      
      !!** Output for check
      !write(80,*) 'amatrx(1,1)=',amatrx(1,1)
      !write(80,*) 'amatrx(2,2)=',amatrx(2,2)
      !write(80,*) 'amatrx(3,3)=',amatrx(3,3)
      !write(80,*) 'amatrx(4,4)=',amatrx(4,4)
      !write(80,*) 'amatrx(5,5)=',amatrx(5,5)
      !write(80,*) 'amatrx(6,6)=',amatrx(6,6)
      !write(80,*) 'amatrx(7,7)=',amatrx(7,7)
      !write(80,*) 'amatrx(8,8)=',amatrx(8,8)
      !write(80,*) 'amatrx(9,9)=',amatrx(9,9)
      !write(80,*) 'amatrx(10,10)=',amatrx(10,10)
      !write(80,*) 'amatrx(11,11)=',amatrx(11,11)
      !write(80,*) 'amatrx(12,12)=',amatrx(12,12)
      
      return
      end subroutine amarhstran